package testkinectv2;

import javax.swing.*;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.Collection;
import java.util.LinkedList;
import java.util.concurrent.*;

public class TestKINECTV2_AudioAndBody extends JFrame implements Runnable, ActionListener, ComponentListener {

	// Kinect color, body and audio
	private long[] sensor = {0};
	private long[] source = {0};
	private long[] reader = {0};
	private long[] descript = {0};
	private long[] mapper = {0};
	private long[] frame = {0};
	private long[] sourceB = {0};
	private long[] readerB = {0};
	private long[] frameB = {0};
	private long[] bodies = new long[H_Kinect.BODY_COUNT];
	private long[] sourceA = {0};
	private long[] readerA = {0};
	private long[] listA = {0};
	private long[] frameA = {0};
	private long[] beamA = {0};
	private long[] subFrameA = {0};
	private long[] correlationA = {0};
	
	private int[] color = null;
	private int width = 0, height = 0;
	private String[] strAudio = new String[H_Kinect.BODY_COUNT];
	private final String AUDIO_TEXT = "SPEAKING";
	
	private AudioBeamFrame[] audioBeamFrame = null;
	
	private class AudioBeamFrame {
		float beamAngle;
		float beamAngleConfidence;
		AudioBeamSubFrame[] subFrame = null;
	}
	private class AudioBeamSubFrame {
		AudioBodyCorrelation[] correlation = null;
	}
	private class AudioBodyCorrelation {
		long bodyTrackingId;
	}
	
	// thread pool
	private ExecutorService threadPool = Executors.newFixedThreadPool(H_Kinect.BODY_COUNT);
	private Collection<Callable<Void>> processes = new LinkedList<Callable<Void>>();

	// wrapper
	private Wrp_KINECTV2 wknct = new Wrp_KINECTV2();
	private Wrp_IKinectSensor wiks = null;
	private Wrp_IColorFrameSource wicfs = null;
	private Wrp_IColorFrameReader wicfr = null;
	private Wrp_IFrameDescription wifd = null;
	private Wrp_ICoordinateMapper wicm = null;
	private Wrp_IColorFrame wicf = null;
	private Wrp_IBodyFrameSource wibfs = null;
	private Wrp_IBodyFrameReader wibfr = null;
	private Wrp_IBodyFrame wibf = null;
	private Wrp_IBody[] wib = new Wrp_IBody[H_Kinect.BODY_COUNT];
	private Wrp_IAudioSource wias = null;
	private Wrp_IAudioBeamFrameReader wiabfr = null;
	private Wrp_IAudioBeamFrameList wiabfl = null;
	private Wrp_IAudioBeamFrame wiabf = null;
	private Wrp_IAudioBeam wiab = null;
	private Wrp_IAudioBeamSubFrame wiabsf = null;
	private Wrp_IAudioBodyCorrelation wiabc = null;
	
	// GUI
	private final int IMG_WIDTH = 1920 / 2;
	private final int IMG_HEIGHT = 1080 / 2;
	private final int BORDER_WIDTH = 4;
	private final int TITLE_HEIGHT = 30;
	private final int PANEL_HEIGHT = 40;
	private final int PANEL_BODY_HEIGHT = 120;
	private JCheckBox chkBody;
	private JButton btnExit;
	private boolean blExit = false; 
	private JPanel pnl;
	private JLabel[] lblAudio = new JLabel[H_Kinect.BODY_COUNT];
	private JPanel pnlAudio;
	private ImageCanvas imgcvs;
	private BufferedImage bfimgOff = null;
	private Graphics2D grphOff = null;
	private boolean blPaint = false;
	private final Color[] colorJoint = {Color.RED, Color.GREEN, Color.BLUE, Color.YELLOW, Color.MAGENTA, Color.CYAN};
	private final Color[] colorText = {new Color(128, 64, 64), new Color(64, 128, 64), new Color(64, 64, 128), 
			new Color(128, 128, 64), new Color(128, 64, 128), new Color(64, 128, 128)};
	private final float LINE_WIDTH = 5.0f;
	
	public static void main(String[] args) {
		TestKINECTV2_AudioAndBody testkinectv2 = new TestKINECTV2_AudioAndBody();
	}
	
	TestKINECTV2_AudioAndBody() {
		this.setTitle("TestKINECTV2_AudioAndBody");
		this.setSize(IMG_WIDTH + BORDER_WIDTH * 2, IMG_HEIGHT + TITLE_HEIGHT + PANEL_BODY_HEIGHT + PANEL_HEIGHT + BORDER_WIDTH);
		this.addComponentListener(this);
		this.setLocationRelativeTo(null);
		this.setLayout(null);
		this.setResizable(true);
		getContentPane().setBackground(Color.DARK_GRAY);
		
		imgcvs = new ImageCanvas();
		imgcvs.setBounds(0, 0, IMG_WIDTH, IMG_HEIGHT);
		getContentPane().add(imgcvs);
		
		pnlAudio = new JPanel();
		pnlAudio.setLayout(new GridLayout(H_Kinect.BODY_COUNT, 1));
		pnlAudio.setBounds(0, IMG_HEIGHT, IMG_WIDTH, PANEL_BODY_HEIGHT);
		
		for (int i = 0; i < lblAudio.length; i++) {
			lblAudio[i] = new JLabel();
			lblAudio[i].setForeground(colorText[i]);
			lblAudio[i].setFont(new Font(Font.MONOSPACED, Font.PLAIN , 12));
			pnlAudio.add(lblAudio[i]);
		}
		
		getContentPane().add(pnlAudio);

		pnl = new JPanel();
		pnl.setLayout(new FlowLayout(FlowLayout.RIGHT, 10, 5));
		pnl.setBounds(0, IMG_HEIGHT + PANEL_BODY_HEIGHT, IMG_WIDTH, PANEL_HEIGHT);
		
		chkBody = new JCheckBox("Body", true);
		pnl.add(chkBody);
		
		btnExit = new JButton("  Exit  ");
		btnExit.addActionListener(this);
		pnl.add(btnExit);
		
		getContentPane().add(pnl);
		
		this.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent ev) {
				System.exit(0);
			}
		});

		this.setVisible(true);
		
		Thread thread = new Thread(this);
		thread.start();
	}
	
	public void actionPerformed(ActionEvent e) {
		JButton btn = (JButton)e.getSource();
		
		if (btn == btnExit) {
			blExit = true;
		}
	}

	public void componentResized(ComponentEvent e) {
		Rectangle rec = e.getComponent().getBounds();
		rec.width -= BORDER_WIDTH * 2;
		rec.height -= TITLE_HEIGHT + PANEL_BODY_HEIGHT + PANEL_HEIGHT + BORDER_WIDTH;
		
		System.out.println("componentResized width: " + Integer.toString(rec.width)
				+ " height: " + Integer.toString(rec.height));

		if ((double)IMG_HEIGHT / (double)IMG_WIDTH >= (double)rec.height / (double)rec.width) {
			int w = (int)((double)rec.height * (double)IMG_WIDTH / (double)IMG_HEIGHT);
			imgcvs.setBounds((rec.width - w) / 2, 0, w, rec.height);
		}
		else {
			int h = (int)((double)rec.width * (double)IMG_HEIGHT / (double)IMG_WIDTH);
			imgcvs.setBounds(0, (rec.height - h) / 2, rec.width, h);
		}
		
		pnlAudio.setBounds(0, rec.height, rec.width, PANEL_BODY_HEIGHT);
		pnl.setBounds(0, rec.height + PANEL_BODY_HEIGHT, rec.width, PANEL_HEIGHT);
	}

	public void componentMoved(ComponentEvent e) {
	}

	public void componentShown(ComponentEvent e) {
	}

	public void componentHidden(ComponentEvent e) {
	}

	private class ImageCanvas extends Canvas {
		public void paint(Graphics g){
			if (blPaint == true) {
				g.drawImage(bfimgOff, 0, 0, imgcvs.getWidth(), imgcvs.getHeight(), this);
				
				blPaint = false;
			}
		}
		
		public void update(Graphics g){
			paint(g);
		}
	}

	public void run() {
		try {
			initialize();
			drawcvs();
		}
		catch (Exception_KINECT e) {
			System.out.println("***** Exception_KINECT: " + e.getMessage());
		}
		finally {
			System.out.println("finally start ");
			try {
				terminate();
			}
			catch (Exception_KINECT e) {
				System.out.println("***** Exception_KINECT: " + e.getMessage());
			}
			System.out.println("finally end ");
		}
		System.exit(0);
	}
	
	private void initialize() throws Exception_KINECT {
		// sensor
		wknct.GetDefaultKinectSensor(sensor);
		System.out.println("sensor: " + Long.toHexString(sensor[0]));
		
		wiks = new Wrp_IKinectSensor(sensor[0]);
		wiks.Open();
		
		// color source
		wiks.get_ColorFrameSource(source);
		System.out.println("source: " + Long.toHexString(source[0]));
		
		wicfs = new Wrp_IColorFrameSource(source[0]);
		
		// color reader
		wicfs.OpenReader(reader);
		System.out.println("reader: " + Long.toHexString(reader[0]));
		
		wicfr = new Wrp_IColorFrameReader(reader[0]);
		
		// color description
		wicfs.get_FrameDescription(descript);
		System.out.println("descript: " + Long.toHexString(descript[0]));
		
		wifd = new Wrp_IFrameDescription(descript[0]);
		
		int[] pWidth = {0};
		int[] pHeight = {0};
		wifd.get_Width(pWidth);
		wifd.get_Height(pHeight);
		System.out.println("width: " + Integer.toString(pWidth[0])
				+ " height: " + Integer.toString(pHeight[0]));
		width = pWidth[0];
		height = pHeight[0];
		
		color = new int[width * height];
		
		// body source
		wiks.get_BodyFrameSource(sourceB);
		System.out.println("sourceB: " + Long.toHexString(sourceB[0]));
		
		wibfs = new Wrp_IBodyFrameSource(sourceB[0]);
		
		// body reader
		wibfs.OpenReader(readerB);
		System.out.println("readerB: " + Long.toHexString(readerB[0]));
		
		wibfr = new Wrp_IBodyFrameReader(readerB[0]);
		
		// audio source
		wiks.get_AudioSource(sourceA);
		System.out.println("sourceA: " + Long.toHexString(sourceA[0]));
		
		wias = new Wrp_IAudioSource(sourceA[0]);
		
		// audio reader
		wias.OpenReader(readerA);
		System.out.println("readerA: " + Long.toHexString(readerA[0]));
		
		wiabfr = new Wrp_IAudioBeamFrameReader(readerA[0]);
		
		// coordinate mapper
		wiks.get_CoordinateMapper(mapper);
		System.out.println("mapper: " + Long.toHexString(mapper[0]));
		
		wicm = new Wrp_ICoordinateMapper(mapper[0]);
		
		
		// initialize array
		for (int i = 0; i < H_Kinect.BODY_COUNT; i++) {
			bodies[i] = 0;
			wib[i] = null;
		}
	}
	
	private void drawcvs() throws Exception_KINECT {
		bfimgOff = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
		grphOff = (Graphics2D)bfimgOff.getGraphics();
		
		for (;;) {
			
			getAudioBodyCorrelation();
			
			if (setColorImage()) {
				blPaint = true;
				imgcvs.repaint();
			}
			
			if (blExit) {
				break;
			}
		}
	}

	private boolean setColorImage() throws Exception_KINECT {
		boolean blSet = false;
		
		// color frame
		frame[0] = 0;
		int hr = wicfr.AcquireLatestFrame(frame);
		System.out.println("frame: " + Long.toHexString(frame[0]));
		
		if (H_WinError.SUCCEEDED(hr) && frame[0] != 0) {
			wicf = new Wrp_IColorFrame(frame[0]);
			
			int capacity = color.length * 4;
			wicf.CopyConvertedFrameDataToArray(capacity, color, H_Kinect.ColorImageFormat_Bgra);
			
			// set color image
			bfimgOff.setRGB(0, 0, width, height, color, 0, width);
			
			// draw bodies
			drawBody();

			blSet = true;
		}
		
		if (wicf != null) wicf.Release();
		wicf = null;
		frame[0] = 0;
		return blSet;
	}
	
	private void getAudioBodyCorrelation() throws Exception_KINECT {
		
		// beam frames
		listA[0] = 0;
		int  hr = wiabfr.AcquireLatestBeamFrames(listA);
		System.out.println("listA: " + Long.toHexString(listA[0]));
		
		if (H_WinError.SUCCEEDED(hr) && listA[0] != 0) {
			wiabfl = new Wrp_IAudioBeamFrameList(listA[0]);
			
			// get beam count
			int[] count = {0};
			wiabfl.get_BeamCount(count);
			System.out.println("count: " + Integer.toString(count[0]));
			
			audioBeamFrame = new AudioBeamFrame[count[0]];
			
			for (int i = 0; i < count[0]; i++) {
				// open beam frame
				frameA[0] = 0;
				wiabfl.OpenAudioBeamFrame(i, frameA);
				System.out.println("frameA#" + Integer.toString(i) + ": " + Long.toHexString(frameA[0]));
				
				wiabf = new Wrp_IAudioBeamFrame(frameA[0]);
				
				// get beam
				beamA[0] = 0;
				wiabf.get_AudioBeam(beamA);
				System.out.println("beamA#" + Integer.toString(i) + ": " + Long.toHexString(beamA[0]));
				
				wiab = new Wrp_IAudioBeam(beamA[0]);
				
				// get beam angle
				float[] angle = {0.0f};
				wiab.get_BeamAngle(angle);
				audioBeamFrame[i] = new AudioBeamFrame();
				audioBeamFrame[i].beamAngle = (float)(angle[0] * 180.0f / Math.PI);
				
				// get beam angle confidence
				float[] confidence = {0.0f};
				wiab.get_BeamAngleConfidence(confidence);
				audioBeamFrame[i].beamAngleConfidence = confidence[0];
				
				// get sub frame count
				int[] subFrameCount = {0};
				wiabf.get_SubFrameCount(subFrameCount);
				System.out.println("subFrameCount#" + Integer.toString(i) + ": " + Integer.toString(subFrameCount[0]));
				
				audioBeamFrame[i].subFrame = new AudioBeamSubFrame[subFrameCount[0]];
				
				for (int j = 0; j < subFrameCount[0]; j++) {
					// get sub frame
					subFrameA[0] = 0;
					wiabf.GetSubFrame(j, subFrameA);
					System.out.println("subFrameA#" + Integer.toString(i) + "," + Integer.toString(j) 
							+ ": " + Long.toHexString(subFrameA[0]));
					
					wiabsf = new Wrp_IAudioBeamSubFrame(subFrameA[0]);
					
					// get correlation count
					int[] correlationCount = {0};
					wiabsf.get_AudioBodyCorrelationCount(correlationCount);
					System.out.println("correlationCount#" + Integer.toString(i) + "," + Integer.toString(j) 
							+ ": " + Integer.toString(correlationCount[0]));
					
					audioBeamFrame[i].subFrame[j] = new AudioBeamSubFrame();
					audioBeamFrame[i].subFrame[j].correlation = new AudioBodyCorrelation[correlationCount[0]];
					
					for (int k = 0; k < correlationCount[0]; k++) {
						// get correlation
						correlationA[0] = 0;
						wiabsf.GetAudioBodyCorrelation(k, correlationA);
						System.out.println("correlationA#" + Integer.toString(i) + "," + Integer.toString(j) 
								 + "," + Integer.toString(k) + ": " + Long.toHexString(correlationA[0]));
						
						wiabc = new Wrp_IAudioBodyCorrelation(correlationA[0]);
						
						// get body tracking id
						long[] trackingId = {0};
						wiabc.get_BodyTrackingId(trackingId);
						System.out.println("trackingId#" + Integer.toString(i) + "," + Integer.toString(j) 
								 + "," + Integer.toString(k) + ": " + Long.toHexString(trackingId[0]));
						
						audioBeamFrame[i].subFrame[j].correlation[k] = new AudioBodyCorrelation();
						audioBeamFrame[i].subFrame[j].correlation[k].bodyTrackingId = trackingId[0];
						
						// release
						if (wiabc != null) wiabc.Release();
						wiabc = null;
						correlationA[0] = 0;
					}
					// release
					if (wiabsf != null) wiabsf.Release();
					wiabsf = null;
					subFrameA[0] = 0;
				}
				
				// release
				if (wiab != null) wiab.Release();
				wiab = null;
				beamA[0] = 0;
				
				if (wiabf != null) wiabf.Release();
				wiabf = null;
				frameA[0] = 0;
			}
		}
		
		// release
		if (wiabfl != null) wiabfl.Release();
		wiabfl = null;
		listA[0] = 0;
	}
	
	private boolean drawBody() throws Exception_KINECT {
		boolean blSet = false;

		// body frame
		frameB[0] = 0;
		int  hr = wibfr.AcquireLatestFrame(frameB);
		System.out.println("frameB: " + Long.toHexString(frameB[0]));
		
		if (H_WinError.SUCCEEDED(hr) && frameB[0] != 0) {
			wibf = new Wrp_IBodyFrame(frameB[0]);
			
			for (int i = 0; i < bodies.length; i++) {
				bodies[i] = 0;
			}
			wibf.GetAndRefreshBodyData(bodies.length, bodies);
			System.out.println("bodies: " + Long.toHexString(bodies[0])
					+ " " + Long.toHexString(bodies[1]) + " " + Long.toHexString(bodies[2])
					+ " " + Long.toHexString(bodies[3]) + " " + Long.toHexString(bodies[4])
					+ " " + Long.toHexString(bodies[5]));
			
			for (int ii = 0; ii < bodies.length; ii++) {
				final int i = ii;
				processes.add(new Callable<Void>() {
			        @Override
			        public Void call() throws Exception_KINECT {
						strAudio[i] = null;
						wib[i] = null;
						
						if (bodies[i] != 0) {
							wib[i] = new Wrp_IBody(bodies[i]);
							
							byte[] tracked = {0};
							wib[i].get_IsTracked(tracked);
							System.out.println("tracked: " + Byte.toString(tracked[0]));
							
							if (tracked[0] != 0) {
								
								Struct_Joint[] joints = new Struct_Joint[H_Kinect.JointType_Count];
								for (int j = 0; j < joints.length; j++) {
									joints[j] = new Struct_Joint();
								}
								wib[i].GetJoints(joints.length, joints);
								
								Struct_ColorSpacePoint[] colorPoints = new Struct_ColorSpacePoint[joints.length];
								for (int j = 0; j < colorPoints.length; j++) {
									if (joints[j].TrackingState != H_Kinect.TrackingState_NotTracked) {
										if (chkFloat(joints[j].Position.X) && chkFloat(joints[j].Position.Y) && chkFloat(joints[j].Position.Z)) {
											colorPoints[j] = new Struct_ColorSpacePoint();
											wicm.MapCameraPointToColorSpace(joints[j].Position, colorPoints[j]);
										}
									}
								}
								
								if (chkBody.isSelected()) {
									// draw joints
									drawJoints(colorPoints, i);
								}
								
								// compare tracking ID
								long[] trackingId = {0};
								wib[i].get_TrackingId(trackingId);
								System.out.println("trackingId: " + Long.toHexString(trackingId[0]));
								
								// draw Audio
								drawAudio(colorPoints, i, trackingId[0]);
							}
						}
						if (strAudio[i] == null) strAudio[i] = "Audio#" + Integer.toString(i) + " NotTracked";
						lblAudio[i].setText(strAudio[i]);
			            return null;
			        }
			    });
						
			}
			
			try {
			    threadPool.invokeAll(processes);
			} catch (InterruptedException e) {
			    throw new RuntimeException(e);
			} finally {
				processes.clear();
			}

			for (int i = 0; i < wib.length; i++) {
				if (wib[i] != null) wib[i].Release();
				wib[i] = null;
				bodies[i] = 0;
			}
			
			blSet = true;
		}
		
		if (wibf != null) wibf.Release();
		wibf = null;
		frameB[0] = 0;
		return blSet;
	}

	private void drawAudio(Struct_ColorSpacePoint[] colorPoints, int bodyNo, long trackingId) throws Exception_KINECT {
		
		if (audioBeamFrame != null) {
			for (int i = 0; i < audioBeamFrame.length; i++) {
				if (audioBeamFrame[i].subFrame != null) {
					for (int j = 0; j < audioBeamFrame[i].subFrame.length; j++) {
						if (audioBeamFrame[i].subFrame[j].correlation != null) {
							for (int k = 0; k < audioBeamFrame[i].subFrame[j].correlation.length; k++) {
								if (audioBeamFrame[i].subFrame[j].correlation[k].bodyTrackingId == trackingId) {
									drawAudioText(AUDIO_TEXT, colorPoints[H_Kinect.JointType_Head], bodyNo);
									
									strAudio[bodyNo] = "Audio#" + Integer.toString(bodyNo) + " Tracked   "
											+ String.format(" BeamAngle:%10.5f", audioBeamFrame[i].beamAngle)
											+ String.format(" BeamAngleConfidence:%10.5f", audioBeamFrame[i].beamAngleConfidence);
									return;
								}
							}
						}
					}
				}
			}
		}
	}

	private void drawAudioText(String str, Struct_ColorSpacePoint colorPoint, int bodyNo) {
		grphOff.setFont(new Font(Font.MONOSPACED, Font.BOLD , 96));
		
		if (colorPoint != null) {
			if (chkFloat(colorPoint.X) && chkFloat(colorPoint.Y)) {
				int pointX = (int)colorPoint.X;
				int pointY = (int)colorPoint.Y;
				
				if (chkRange(pointX, pointY)) {
					grphOff.setColor(Color.WHITE);
					grphOff.drawString(str, colorPoint.X + 2, colorPoint.Y);
					grphOff.drawString(str, colorPoint.X - 2, colorPoint.Y);
					grphOff.drawString(str, colorPoint.X, colorPoint.Y + 2);
					grphOff.drawString(str, colorPoint.X, colorPoint.Y - 2);
					grphOff.setColor(colorJoint[bodyNo]);
					grphOff.drawString(str, colorPoint.X, colorPoint.Y);
				}
			}
		}
	}
	
	private boolean chkFloat(float f) {
		if (Float.isNaN(f)) return false;
		if (Float.isInfinite(f)) return false;
		return true;
	}
	
	private boolean chkRange(int pointX, int pointY) {
		if (pointX < 0 || width <= pointX) return false;
		if (pointY < 0 || height <= pointY) return false;
		return true;
	}
	
	private boolean setPolyline(Struct_ColorSpacePoint[] colorPoints, int[] pointX, int[] pointY) {
		
		for (int i = 0; i < colorPoints.length; i++) {
			if (colorPoints[i] == null) return false;
			if (!chkFloat(colorPoints[i].X) || !chkFloat(colorPoints[i].Y)) return false;
			pointX[i] = (int)colorPoints[i].X;
			pointY[i] = (int)colorPoints[i].Y;
			if (!chkRange(pointX[i], pointY[i])) return false;
		}
	
		return true;
	}
	
	private void drawJoints(Struct_ColorSpacePoint[] colorPoints, int bodyNo) {
		
		grphOff.setStroke(new BasicStroke(LINE_WIDTH));
		grphOff.setColor(colorJoint[bodyNo]);
		final int r = 5;
		final int r2 = r * 2;
		
		// draw joints
		for (int i = 0; i < colorPoints.length; i++) {
			if (colorPoints[i] != null) {
				if (chkFloat(colorPoints[i].X) && chkFloat(colorPoints[i].Y)) {
					int pointX = (int)colorPoints[i].X;
					int pointY = (int)colorPoints[i].Y;
					
					if (chkRange(pointX, pointY)) {
						grphOff.drawOval(pointX - r, pointY - r, r2, r2);
					}
				}
			}
		}
		
		// draw lines
		int[] pointx = new int[6];
		int[] pointy = new int[6];
		
		grphOff.setStroke(new BasicStroke(LINE_WIDTH));
		grphOff.setColor(colorJoint[bodyNo]);

		
		final Struct_ColorSpacePoint[] headSpine = {
				colorPoints[H_Kinect.JointType_Head],
				colorPoints[H_Kinect.JointType_Neck],
				colorPoints[H_Kinect.JointType_SpineShoulder],
				colorPoints[H_Kinect.JointType_SpineMid],
				colorPoints[H_Kinect.JointType_SpineBase]};
		
		if (setPolyline(headSpine, pointx, pointy)) {
			grphOff.drawPolyline(pointx, pointy, headSpine.length);
		}
		
		final Struct_ColorSpacePoint[] handRight = {
				colorPoints[H_Kinect.JointType_SpineShoulder],
				colorPoints[H_Kinect.JointType_ShoulderRight],
				colorPoints[H_Kinect.JointType_ElbowRight],
				colorPoints[H_Kinect.JointType_WristRight],
				colorPoints[H_Kinect.JointType_HandRight],
				colorPoints[H_Kinect.JointType_HandTipRight]};
		
		if (setPolyline(handRight, pointx, pointy)) {
			grphOff.drawPolyline(pointx, pointy, handRight.length);
		}
		
		final Struct_ColorSpacePoint[] thumbRight = {
				colorPoints[H_Kinect.JointType_HandRight],
				colorPoints[H_Kinect.JointType_ThumbRight]};
		
		if (setPolyline(thumbRight, pointx, pointy)) {
			grphOff.drawPolyline(pointx, pointy, thumbRight.length);
		}
		
		final Struct_ColorSpacePoint[] handLeft = {
				colorPoints[H_Kinect.JointType_SpineShoulder],
				colorPoints[H_Kinect.JointType_ShoulderLeft],
				colorPoints[H_Kinect.JointType_ElbowLeft],
				colorPoints[H_Kinect.JointType_WristLeft],
				colorPoints[H_Kinect.JointType_HandLeft],
				colorPoints[H_Kinect.JointType_HandTipLeft]};
		
		if (setPolyline(handLeft, pointx, pointy)) {
			grphOff.drawPolyline(pointx, pointy, handLeft.length);
		}
		
		final Struct_ColorSpacePoint[] thumbLeft = {
				colorPoints[H_Kinect.JointType_HandLeft],
				colorPoints[H_Kinect.JointType_ThumbLeft]};
		
		if (setPolyline(thumbLeft, pointx, pointy)) {
			grphOff.drawPolyline(pointx, pointy, thumbLeft.length);
		}
		
		final Struct_ColorSpacePoint[] footRight = {
				colorPoints[H_Kinect.JointType_SpineBase],
				colorPoints[H_Kinect.JointType_HipRight],
				colorPoints[H_Kinect.JointType_KneeRight],
				colorPoints[H_Kinect.JointType_AnkleRight],
				colorPoints[H_Kinect.JointType_FootRight]};
		
		if (setPolyline(footRight, pointx, pointy)) {
			grphOff.drawPolyline(pointx, pointy, footRight.length);
		}
		
		final Struct_ColorSpacePoint[] footLeft = {
				colorPoints[H_Kinect.JointType_SpineBase],
				colorPoints[H_Kinect.JointType_HipLeft],
				colorPoints[H_Kinect.JointType_KneeLeft],
				colorPoints[H_Kinect.JointType_AnkleLeft],
				colorPoints[H_Kinect.JointType_FootLeft]};
		
		if (setPolyline(footLeft, pointx, pointy)) {
			grphOff.drawPolyline(pointx, pointy, footLeft.length);
		}
		
	}

	private void terminate() throws Exception_KINECT {
		if (wiabc != null) wiabc.Release();
		if (wiabsf != null) wiabsf.Release();
		if (wiab != null) wiab.Release();
		if (wiabf != null) wiabf.Release();
		if (wiabfl != null) wiabfl.Release();
		if (wiabfr != null) wiabfr.Release();
		if (wias != null) wias.Release();
		
		for (int i = 0; i < wib.length; i++) {
			if (wib[i] != null) wib[i].Release();
		}
		
		if (wicf != null) wicf.Release();
		if (wicfr != null) wicfr.Release();
		if (wifd != null) wifd.Release();
		if (wicm != null) wicm.Release();
		if (wicfs != null) wicfs.Release();
		if (wibf != null) wibf.Release();
		if (wibfr != null) wibfr.Release();
		if (wibfs != null) wibfs.Release();
		
		if (wiks != null) {
			wiks.Close();
			wiks.Release();
		}
		
		if (grphOff != null) grphOff.dispose();
		threadPool.shutdown();
	}

}
