package testkinectv2;

import javax.swing.*;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.*;

public class TestKINECTV2_Controls extends JFrame implements Runnable, ActionListener {

	// Kinect core window
	private long[] sensor = {0};
	private long[] coreWin = {0};
	private long[] eventArgs = {0};
	private long[] point = {0};
	private long[] prop = {0};
	private long[] hMoved = {H_WinBase.INVALID_HANDLE_VALUE};
	private long[] hEntered = {H_WinBase.INVALID_HANDLE_VALUE};
	private long[] hExited = {H_WinBase.INVALID_HANDLE_VALUE};
	private long[] hEvents = null;
	
	private int[] color = null;
	private int width = 0, height = 0;
	private Queue<MovedPoint> queue = new LinkedList<MovedPoint>();
	private final int QUEUE_MAX_SIZE = 100;
	private final long DRAW_TIME_OVER = 3000L;
	
	private class MovedPoint {
		Struct_PointF pointf;
		int handType;
		long time;
	}
	
	// wrapper
	private Wrp_KINECTV2 wknct = new Wrp_KINECTV2();
	private Wrp_Kernel32 wkrnl = new Wrp_Kernel32();
	private Wrp_IKinectSensor wiks = null;
	private Wrp_IKinectCoreWindow wikcw = null;
	private Wrp_IKinectPointerEventArgs wikpea = null;
	private Wrp_IKinectPointerPoint wikpp = null;
	private Wrp_IKinectPointerPointProperties wikppp = null;
	
	// GUI
	private final int IMG_WIDTH = 960;
	private final int IMG_HEIGHT = 540;
	private final int BORDER_WIDTH = 4;
	private final int TITLE_HEIGHT = 30;
	private final int PANEL_HEIGHT = 40;
	private JButton btnExit;
	private boolean blExit = false; 
	private JPanel pnl;
	private ImageCanvas imgcvs;
	private BufferedImage bfimgOff = new BufferedImage(IMG_WIDTH, IMG_HEIGHT, BufferedImage.TYPE_INT_RGB);
	private Graphics2D grphOff = (Graphics2D)bfimgOff.getGraphics();
	private boolean blPaint = false;
	private final Color[] colorMoved = {Color.RED, Color.GREEN, Color.BLUE, Color.YELLOW, Color.MAGENTA, Color.CYAN};
	private final float LINE_WIDTH = 5.0f;
	
	public static void main(String[] args) {
		TestKINECTV2_Controls testkinectv2 = new TestKINECTV2_Controls();
	}
	
	TestKINECTV2_Controls() {
		this.setTitle("TestKINECTV2_Controls");
		this.setSize(IMG_WIDTH + BORDER_WIDTH * 2, IMG_HEIGHT + TITLE_HEIGHT + PANEL_HEIGHT + BORDER_WIDTH);
		this.setLocationRelativeTo(null);
		this.setLayout(null);
		this.setResizable(false);
		
		imgcvs = new ImageCanvas();
		imgcvs.setBounds(0, 0, IMG_WIDTH, IMG_HEIGHT);
		getContentPane().add(imgcvs);
		
		pnl = new JPanel();
		pnl.setLayout(new FlowLayout(FlowLayout.RIGHT, 10, 5));
		pnl.setBounds(0, IMG_HEIGHT, IMG_WIDTH, PANEL_HEIGHT);
		
		btnExit = new JButton("  Exit  ");
		btnExit.addActionListener(this);
		pnl.add(btnExit);
		
		getContentPane().add(pnl);
		
		this.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent ev) {
				System.exit(0);
			}
		});

		this.setVisible(true);
		
		Thread thread = new Thread(this);
		thread.start();
	}

	public void actionPerformed(ActionEvent e) {
		JButton btn = (JButton)e.getSource();
		
		if (btn == btnExit) {
			blExit = true;
		}
	}

	private class ImageCanvas extends Canvas {
		public void paint(Graphics g){
			if (blPaint == true) {
				g.drawImage(bfimgOff, 0, 0, this);
				blPaint = false;
			}
		}
		
		public void update(Graphics g){
			paint(g);
		}
	}
	
	public void run() {
		try {
			initialize();
			drawcvs();
		}
		catch (Exception_KINECT e) {
			System.out.println("***** Exception_KINECT: " + e.getMessage());
		}
		catch (Exception_Win64 e) {
			System.out.println("***** Exception_Win64: " + e.getMessage());
		}
		finally {
			System.out.println("finally start ");
			try {
				terminate();
			}
			catch (Exception_KINECT e) {
				System.out.println("***** Exception_KINECT: " + e.getMessage());
			}
			catch (Exception_Win64 e) {
				System.out.println("***** Exception_Win64: " + e.getMessage());
			}
			System.out.println("finally end ");
		}
		System.exit(0);
	}

	private void initialize() throws Exception_KINECT {
		// sensor
		wknct.GetDefaultKinectSensor(sensor);
		System.out.println("sensor: " + Long.toHexString(sensor[0]));
		
		wiks = new Wrp_IKinectSensor(sensor[0]);
		wiks.Open();
		
		// core window
		wknct.GetKinectCoreWindowForCurrentThread(coreWin);
		System.out.println("coreWin: " + Long.toHexString(coreWin[0]));
		
		wikcw = new Wrp_IKinectCoreWindow(coreWin[0]);
		
		// subscribe pointer moved
		wikcw.SubscribePointerMoved(hMoved);
		System.out.println("hMoved: " + Long.toHexString(hMoved[0]));
		
		// subscribe pointer entered
		wikcw.SubscribePointerEntered(hEntered);
		System.out.println("hEntered: " + Long.toHexString(hEntered[0]));
		
		// subscribe pointer exited
		wikcw.SubscribePointerExited(hExited);
		System.out.println("hExited: " + Long.toHexString(hExited[0]));

		width = IMG_WIDTH;
		height = IMG_HEIGHT;
		color = new int[width * height];
		
		hEvents = new long[]{hMoved[0], hEntered[0], hExited[0]};
		
	}
	
	private void drawcvs() throws Exception_KINECT, Exception_Win64 {
		
		for (;;) {
			
			getMovedPoint();
			
			// set color image
			bfimgOff.setRGB(0, 0, width, height, color, 0, width);  // black background

			drawMovedPoint();
			
			blPaint = true;
			imgcvs.repaint();
			
			if (blExit) {
				break;
			}
		}
	}
	
	private boolean getMovedPoint() throws Exception_KINECT, Exception_Win64 {
		boolean blSet = false;
		final int WAIT_OBJECT_0 = H_WinBase.WAIT_OBJECT_0;
		final int WAIT_OBJECT_1 = H_WinBase.WAIT_OBJECT_0 + 1;
		final int WAIT_OBJECT_2 = H_WinBase.WAIT_OBJECT_0 + 2;
		
		int objectNo = wkrnl.WaitForMultipleObjects(hEvents.length, hEvents, false, 1000);
		System.out.println("objectNo: " + Integer.toHexString(objectNo));
		
		switch (objectNo) {
		case WAIT_OBJECT_0:
			// get moved event data
			wikcw.GetPointerMovedEventData(hMoved[0], eventArgs);
			System.out.println("eventArgs: " + Long.toHexString(eventArgs[0]));
			
			wikpea = new Wrp_IKinectPointerEventArgs(eventArgs[0]);
			
			// get current point
			wikpea.get_CurrentPoint(point);
			System.out.println("point: " + Long.toHexString(point[0]));
			
			if (point[0] != 0) {
				wikpp = new Wrp_IKinectPointerPoint(point[0]);
				
				// get properties
				wikpp.get_Properties(prop);
				System.out.println("prop: " + Long.toHexString(prop[0]));
				
				wikppp = new Wrp_IKinectPointerPointProperties(prop[0]);
				
				byte[] isInRange = {0};
				wikppp.get_IsInRange(isInRange);
				System.out.println("isInRange: " + Byte.toString(isInRange[0]));
				
				if (isInRange[0] != 0) {
					// get hand type
					int[] handType = {0};
					wikppp.get_HandType(handType);
					System.out.println("handType: " + Integer.toString(handType[0]));
					
					// get position
					Struct_PointF pointf = new Struct_PointF();
					wikpp.get_Position(pointf);
					
					if (chkFloat(pointf.X) && chkFloat(pointf.Y)) {
						if (handType[0] == H_Kinect.HandType_LEFT || handType[0] == H_Kinect.HandType_RIGHT) {
							MovedPoint movedPoint = new MovedPoint();
							movedPoint.pointf = pointf;
							movedPoint.handType = handType[0];
							movedPoint.time = System.currentTimeMillis();
							while (queue.size() >= QUEUE_MAX_SIZE) queue.poll();
							queue.offer(movedPoint);
							blSet = true;
						}
					}
				}
			}
			
			if (wikppp != null) wikppp.Release();
			wikppp = null;
			prop[0] = 0;
			if (wikpp != null) wikpp.Release();
			wikpp = null;
			point[0] = 0;
			if (wikpea != null) wikpea.Release();
			wikpea = null;
			eventArgs[0] = 0;
			
			break;
		case WAIT_OBJECT_1:
			
			break;
		case WAIT_OBJECT_2:
			
			break;
			
		}
		return blSet;
	}
	
	private void drawMovedPoint() {
		
		if (queue.isEmpty()) return;
		while (queue.size() > QUEUE_MAX_SIZE) queue.poll();
		while ((System.currentTimeMillis() - queue.peek().time) > DRAW_TIME_OVER) {
			queue.poll();
			if (queue.isEmpty()) return;
		}
		
		grphOff.setStroke(new BasicStroke(LINE_WIDTH));
		final int r = 5;
		final int r2 = r * 2;
		
		Iterator<MovedPoint> itr = queue.iterator();

		while(itr.hasNext()) {
			MovedPoint movedPoint = (MovedPoint)itr.next();
			if (movedPoint.handType == H_Kinect.HandType_LEFT) {
				grphOff.setColor(colorMoved[0]);
				grphOff.fillOval((int)(movedPoint.pointf.X * width) - r, (int)(movedPoint.pointf.Y * height) - r, r2, r2);
			}
			else if (movedPoint.handType == H_Kinect.HandType_RIGHT) {
				grphOff.setColor(colorMoved[1]);
				grphOff.fillRect((int)(movedPoint.pointf.X * width) - r, (int)(movedPoint.pointf.Y * height) - r, r2, r2);
			}
		}		
	}
	
	private boolean chkFloat(float f) {
		if (Float.isNaN(f)) return false;
		if (Float.isInfinite(f)) return false;
		if (f < 0.0f || 1.0f < f) return false;
		return true;
	}
	
	private void terminate() throws Exception_KINECT, Exception_Win64 {
		if (wikppp != null) wikppp.Release();
		if (wikpp != null) wikpp.Release();
		if (wikpea != null) wikpea.Release();
		
		if (wikcw != null) {
			wikcw.UnsubscribePointerMoved(hMoved[0]);
			wikcw.UnsubscribePointerEntered(hEntered[0]);
			wikcw.UnsubscribePointerExited(hExited[0]);
			wikcw.Release();
		}
		
		if (wiks != null) {
			wiks.Close();
			wiks.Release();
		}
		
		if (grphOff != null) grphOff.dispose();
	}

}
