package testkinectv2;

import javax.swing.*;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.Collection;
import java.util.LinkedList;
import java.util.concurrent.*;

public class TestKINECTV2_CoordinateMapper extends JFrame implements Runnable, ActionListener, ComponentListener {

	// Kinect color, depth and body index 
	private long[] sensor = {0};
	private long[] source = {0};
	private long[] reader = {0};
	private long[] descript = {0};
	private long[] mapper = {0};
	private long[] frame = {0};
	private long[] sourceD = {0};
	private long[] readerD = {0};
	private long[] descriptD = {0};
	private long[] frameD = {0};
	private long[] sourceBI = {0};
	private long[] readerBI = {0};
	private long[] descriptBI = {0};
	private long[] frameBI = {0};
	
	private int[] color = null;
	private int[] colorBack = null;
	private int[] colorMerge = null;

	private int width = 0, height = 0;
	private int widthD = 0, heightD = 0;
	private int widthBI = 0, heightBI = 0;
	private final String BACKGROUND_IMAGE_FILE = "backimgv2.jpg";
	private final int BACK_IMG_WIDTH = 1920;
	private final int BACK_IMG_HEIGHT = 1080;
	
	// thread pool
	private ExecutorService threadPool = Executors.newFixedThreadPool(8);
	private Collection<Callable<Void>> processes = new LinkedList<Callable<Void>>();
	private final int COLOR_INDEX_STEP = 64800;
	
	// wrapper
	private Wrp_KINECTV2 wknct = new Wrp_KINECTV2();
	private Wrp_IKinectSensor wiks = null;
	private Wrp_IColorFrameSource wicfs = null;
	private Wrp_IColorFrameReader wicfr = null;
	private Wrp_IFrameDescription wifd = null;
	private Wrp_ICoordinateMapper wicm = null;
	private Wrp_IColorFrame wicf = null;
	private Wrp_IDepthFrameSource widfs = null;
	private Wrp_IDepthFrameReader widfr = null;
	private Wrp_IFrameDescription wifdd = null;
	private Wrp_IDepthFrame widf = null;
	private Wrp_IBodyIndexFrameSource wibifs = null;
	private Wrp_IBodyIndexFrameReader wibifr = null;
	private Wrp_IFrameDescription wifdbi = null;
	private Wrp_IBodyIndexFrame wibif = null;

	// GUI
	private final int IMG_WIDTH = 1920 / 2;
	private final int IMG_HEIGHT = 1080 / 2;
	private final int BORDER_WIDTH = 4;
	private final int TITLE_HEIGHT = 30;
	private final int PANEL_HEIGHT = 40;
	private JButton btnExit;
	private boolean blExit = false; 
	private JPanel pnl;
	private ImageCanvas imgcvs;
	private BufferedImage bfimgOff = null;
	private Graphics2D grphOff = null;
	private boolean blPaint = false;
	
	
	public static void main(String[] args) {
		TestKINECTV2_CoordinateMapper testkinectv2 = new TestKINECTV2_CoordinateMapper();
	}
	
	TestKINECTV2_CoordinateMapper() {
		this.setTitle("TestKINECTV2_CoordinateMapper");
		this.setSize(IMG_WIDTH + BORDER_WIDTH * 2, IMG_HEIGHT + TITLE_HEIGHT + PANEL_HEIGHT + BORDER_WIDTH);
		this.addComponentListener(this);
		this.setLocationRelativeTo(null);
		this.setLayout(null);
		this.setResizable(true);
		getContentPane().setBackground(Color.DARK_GRAY);
		
		imgcvs = new ImageCanvas();
		imgcvs.setBounds(0, 0, IMG_WIDTH, IMG_HEIGHT);
		getContentPane().add(imgcvs);
		
		pnl = new JPanel();
		pnl.setLayout(new FlowLayout(FlowLayout.RIGHT, 10, 5));
		pnl.setBounds(0, IMG_HEIGHT, IMG_WIDTH, PANEL_HEIGHT);
		
		btnExit = new JButton("  Exit  ");
		btnExit.addActionListener(this);
		pnl.add(btnExit);
		
		getContentPane().add(pnl);
		
		this.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent ev) {
				System.exit(0);
			}
		});

		this.setVisible(true);
		
		Thread thread = new Thread(this);
		thread.start();
	}

	public void actionPerformed(ActionEvent e) {
		JButton btn = (JButton)e.getSource();
		
		if (btn == btnExit) {
			blExit = true;
		}
	}

	public void componentResized(ComponentEvent e) {
		Rectangle rec = e.getComponent().getBounds();
		rec.width -= BORDER_WIDTH * 2;
		rec.height -= TITLE_HEIGHT + PANEL_HEIGHT + BORDER_WIDTH;
		
		System.out.println("componentResized width: " + Integer.toString(rec.width)
				+ " height: " + Integer.toString(rec.height));

		if ((double)IMG_HEIGHT / (double)IMG_WIDTH >= (double)rec.height / (double)rec.width) {
			int w = (int)((double)rec.height * (double)IMG_WIDTH / (double)IMG_HEIGHT);
			imgcvs.setBounds((rec.width - w) / 2, 0, w, rec.height);
		}
		else {
			int h = (int)((double)rec.width * (double)IMG_HEIGHT / (double)IMG_WIDTH);
			imgcvs.setBounds(0, (rec.height - h) / 2, rec.width, h);
		}
		
		pnl.setBounds(0, rec.height, rec.width, PANEL_HEIGHT);
	}

	public void componentMoved(ComponentEvent e) {
	}

	public void componentShown(ComponentEvent e) {
	}

	public void componentHidden(ComponentEvent e) {
	}

	private class ImageCanvas extends Canvas {
		public void paint(Graphics g){
			if (blPaint == true) {
				g.drawImage(bfimgOff, 0, 0, imgcvs.getWidth(), imgcvs.getHeight(), this);
				
				blPaint = false;
			}
		}
		
		public void update(Graphics g){
			paint(g);
		}
	}

	public void run() {
		try {
			initialize();
			drawcvs();
		}
		catch (Exception_KINECT e) {
			System.out.println("***** Exception_KINECT: " + e.getMessage());
		}
		finally {
			System.out.println("finally start ");
			try {
				terminate();
			}
			catch (Exception_KINECT e) {
				System.out.println("***** Exception_KINECT: " + e.getMessage());
			}
			System.out.println("finally end ");
		}
		System.exit(0);
	}
	
	private void initialize() throws Exception_KINECT {
		// sensor
		wknct.GetDefaultKinectSensor(sensor);
		System.out.println("sensor: " + Long.toHexString(sensor[0]));
		
		wiks = new Wrp_IKinectSensor(sensor[0]);
		wiks.Open();
		
		// color source
		wiks.get_ColorFrameSource(source);
		System.out.println("source: " + Long.toHexString(source[0]));
		
		wicfs = new Wrp_IColorFrameSource(source[0]);
		
		// color reader
		wicfs.OpenReader(reader);
		System.out.println("reader: " + Long.toHexString(reader[0]));
		
		wicfr = new Wrp_IColorFrameReader(reader[0]);
		
		// color description
		wicfs.get_FrameDescription(descript);
		System.out.println("descript: " + Long.toHexString(descript[0]));
		
		wifd = new Wrp_IFrameDescription(descript[0]);
		
		int[] pWidth = {0};
		int[] pHeight = {0};
		wifd.get_Width(pWidth);
		wifd.get_Height(pHeight);
		System.out.println("width: " + Integer.toString(pWidth[0])
				+ " height: " + Integer.toString(pHeight[0]));
		width = pWidth[0];
		height = pHeight[0];
		
		color = new int[width * height];
		colorMerge = new int[width * height];
		
		// depth source
		wiks.get_DepthFrameSource(sourceD);
		System.out.println("sourceD: " + Long.toHexString(sourceD[0]));
		
		widfs = new Wrp_IDepthFrameSource(sourceD[0]);
		
		// depth reader
		widfs.OpenReader(readerD);
		System.out.println("readerD: " + Long.toHexString(readerD[0]));
		
		widfr = new Wrp_IDepthFrameReader(readerD[0]);

		// depth description
		widfs.get_FrameDescription(descriptD);
		System.out.println("descriptD: " + Long.toHexString(descriptD[0]));
		
		wifdd = new Wrp_IFrameDescription(descriptD[0]);
		
		int[] pWidthD = {0};
		int[] pHeightD = {0};
		wifdd.get_Width(pWidthD);
		wifdd.get_Height(pHeightD);
		System.out.println("widthD: " + Integer.toString(pWidthD[0])
				+ " heightD: " + Integer.toString(pHeightD[0]));
		widthD = pWidthD[0];
		heightD = pHeightD[0];
		
		// body index source
		wiks.get_BodyIndexFrameSource(sourceBI);
		System.out.println("sourceBI: " + Long.toHexString(sourceBI[0]));
		
		wibifs = new Wrp_IBodyIndexFrameSource(sourceBI[0]);
		
		// body index reader
		wibifs.OpenReader(readerBI);
		System.out.println("readerBI: " + Long.toHexString(readerBI[0]));
		
		wibifr = new Wrp_IBodyIndexFrameReader(readerBI[0]);
		
		// body index description
		wibifs.get_FrameDescription(descriptBI);
		System.out.println("descriptBI: " + Long.toHexString(descriptBI[0]));
		
		wifdbi = new Wrp_IFrameDescription(descriptBI[0]);
		
		int[] pWidthBI = {0};
		int[] pHeightBI = {0};
		wifdbi.get_Width(pWidthBI);
		wifdbi.get_Height(pHeightBI);
		System.out.println("widthBI: " + Integer.toString(pWidthBI[0])
				+ " heightBI: " + Integer.toString(pHeightBI[0]));
		widthBI = pWidthBI[0];
		heightBI = pHeightBI[0];
		
		// coordinate mapper
		wiks.get_CoordinateMapper(mapper);
		System.out.println("mapper: " + Long.toHexString(mapper[0]));
		
		wicm = new Wrp_ICoordinateMapper(mapper[0]);
		
		// read background image 
		colorBack = new int[BACK_IMG_WIDTH * BACK_IMG_HEIGHT];
		readBackImage();
	}
	
	private void readBackImage() {
		Image backImage = Toolkit.getDefaultToolkit().getImage(BACKGROUND_IMAGE_FILE);
        PixelGrabber pg = new PixelGrabber(backImage, 0, 0, BACK_IMG_WIDTH, BACK_IMG_HEIGHT, colorBack, 0, BACK_IMG_WIDTH);
        try {
           pg.grabPixels();
        } catch(InterruptedException ie) {}
	}

	private void drawcvs() throws Exception_KINECT {
		bfimgOff = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
		grphOff = (Graphics2D)bfimgOff.getGraphics();
		
		for (;;) {
			
			if (setColorImage()) {
				blPaint = true;
				imgcvs.repaint();
			}
			
			if (blExit) {
				break;
			}
		}
	}
	
	private boolean setColorImage() throws Exception_KINECT {
		boolean blSet = false;
		
		// color frame
		int hr = wicfr.AcquireLatestFrame(frame);
		System.out.println("frame: " + Long.toHexString(frame[0]));
		
		if (H_WinError.SUCCEEDED(hr) && frame[0] != 0) {
			wicf = new Wrp_IColorFrame(frame[0]);
			
			int capacity = color.length * 4;
			wicf.CopyConvertedFrameDataToArray(capacity, color, H_Kinect.ColorImageFormat_Bgra);
			
			// get depth data
			short[] depthData = getDepthData();
			
			// get body index data
			byte[] bodyIndexData = getBodyIndexData();
			
			if (depthData != null && bodyIndexData != null) {
				// get depth points for color frame
				float[] depthPoints = new float[color.length * 2]; // DepthSpacePoint {float X, float Y;}
				wicm.MapColorFrameToDepthSpace(depthData.length, depthData, color.length, depthPoints);
				
				for (int ii = 0; ii < color.length; ii += COLOR_INDEX_STEP) {
					final int start = ii;
					final int end = start + COLOR_INDEX_STEP;
					final float[] depthPointsFinal = depthPoints;
					final byte[] bodyIndexDataFinal = bodyIndexData;
					
					processes.add(new Callable<Void>() {
				        @Override
				        public Void call() throws Exception_KINECT {
					
							for (int i = start; i < end; i++) {
								if (chkFloat(depthPointsFinal[i * 2]) && chkFloat(depthPointsFinal[i * 2 + 1])) {
									int pointX = (int)depthPointsFinal[i * 2];
									int pointY = (int)depthPointsFinal[i * 2 + 1];
									
									if (chkRange(pointX, pointY)) {
										int index = pointX + pointY * widthBI;
										if (0 <= bodyIndexDataFinal[index] && bodyIndexDataFinal[index] <= 5) {
											colorMerge[i] = color[i];
											continue;
										}
									}
								}
								colorMerge[i] = colorBack[i];
							}
				            return null;
				        }
				    });
							
				}
				
				try {
				    threadPool.invokeAll(processes);
				} catch (InterruptedException e) {
				    throw new RuntimeException(e);
				} finally {
					processes.clear();
				}

				// set color image
				bfimgOff.setRGB(0, 0, width, height, colorMerge, 0, width);

				blSet = true;
			}
		}
		
		if (wicf != null) wicf.Release();
		wicf = null;
		frame[0] = 0;
		return blSet;
	}

	private short[] getDepthData() throws Exception_KINECT {
		short[] depthData = null;
		
		// depth frame
		int hr = widfr.AcquireLatestFrame(frameD);
		System.out.println("frameD: " + Long.toHexString(frameD[0]));
		
		if (H_WinError.SUCCEEDED(hr) && frameD[0] != 0) {
			widf = new Wrp_IDepthFrame(frameD[0]);
			
			int[] capacity = {0};
			depthData = widf.AccessUnderlyingBuffer(capacity);
			System.out.println("AccessUnderlyingBuffer capacity: " + Integer.toString(capacity[0])
					+ " buffer.length: " + Integer.toString(depthData.length));
			
		}
		
		if (widf != null) widf.Release();
		widf = null;
		frameD[0] = 0;
		return depthData;
	}
	
	private byte[] getBodyIndexData() throws Exception_KINECT {
		byte[] bodyIndexData = null;

		// body index frame
		int  hr = wibifr.AcquireLatestFrame(frameBI);
		System.out.println("frameBI: " + Long.toHexString(frameBI[0]));
		
		if (H_WinError.SUCCEEDED(hr) && frameBI[0] != 0) {
			wibif = new Wrp_IBodyIndexFrame(frameBI[0]);
			
			int[] capacity = {0};
			bodyIndexData = wibif.AccessUnderlyingBuffer(capacity);
			System.out.println("AccessUnderlyingBuffer capacity: " + Integer.toString(capacity[0])
					+ " buffer.length: " + Integer.toString(bodyIndexData.length));

		}
		
		if (wibif != null) wibif.Release();
		wibif = null;
		frameBI[0] = 0;
		return bodyIndexData;
	}

	private boolean chkFloat(float f) {
		if (Float.isNaN(f)) return false;
		if (Float.isInfinite(f)) return false;
		return true;
	}
	
	private boolean chkRange(int pointX, int pointY) {
		if (pointX < 0 || widthBI <= pointX) return false;
		if (pointY < 0 || heightBI <= pointY) return false;
		return true;
	}

	void terminate() throws Exception_KINECT {
		if (wicf != null) wicf.Release();
		if (wicfr != null) wicfr.Release();
		if (wifd != null) wifd.Release();
		if (wicfs != null) wicfs.Release();
		if (wicm != null) wicm.Release();
		if (widf != null) widf.Release();
		if (widfr != null) widfr.Release();
		if (wifdd != null) wifdd.Release();
		if (widfs != null) widfs.Release();
		if (wibif != null) wibif.Release();
		if (wibifr != null) wibifr.Release();
		if (wifdbi != null) wifdbi.Release();
		if (wibifs != null) wibifs.Release();
		
		if (wiks != null) {
			wiks.Close();
			wiks.Release();
		}
		
		if (grphOff != null) grphOff.dispose();
		threadPool.shutdown();
	}

}
