package testkinectv2;

import javax.swing.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;

public class TestKINECTV2_DepthAndBodyIndex extends JFrame implements Runnable, ActionListener {
	
	// Kinect depth and body index
	private long[] sensor = {0};
	private long[] source = {0};
	private long[] reader = {0};
	private long[] descript = {0};
	private long[] frame = {0};
	private long[] sourceBI = {0};
	private long[] readerBI = {0};
	private long[] descriptBI = {0};
	private long[] frameBI = {0};
	
	private int[] color = null;
	private int[] colorBI = null;
	private int width = 0, height = 0;
	private int widthBI = 0, heightBI = 0;
	private int distanceMin = 0, distanceMax = 0;
	private final int[] colorTable = {0x00ff0000, 0x0000ff00, 0x000000ff, 0x00ffff00, 0x00ff00ff, 0x0000ffff};
	
	// wrapper
	private Wrp_KINECTV2 wknct = new Wrp_KINECTV2();
	private Wrp_IKinectSensor wiks = null;
	private Wrp_IDepthFrameSource widfs = null;
	private Wrp_IDepthFrameReader widfr = null;
	private Wrp_IFrameDescription wifd = null;
	private Wrp_IDepthFrame widf = null;
	private Wrp_IBodyIndexFrameSource wibifs = null;
	private Wrp_IBodyIndexFrameReader wibifr = null;
	private Wrp_IFrameDescription wifdbi = null;
	private Wrp_IBodyIndexFrame wibif = null;
	
	// GUI
	private final int IMG_WIDTH = 512;
	private final int IMG_HEIGHT = 424;
	private final int BORDER_WIDTH = 4;
	private final int TITLE_HEIGHT = 30;
	private final int PANEL_HEIGHT = 40;
	private JButton btnExit;
	private boolean blExit = false; 
	private JPanel pnl;
	private ImageCanvas imgcvs;
	private BufferedImage bfimgOff = new BufferedImage(IMG_WIDTH, IMG_HEIGHT, BufferedImage.TYPE_INT_RGB);
	private Graphics grphOff = bfimgOff.getGraphics();
	private boolean blPaint = false;
	private ImageCanvasBI imgcvsBI;
	private BufferedImage bfimgOffBI = new BufferedImage(IMG_WIDTH, IMG_HEIGHT, BufferedImage.TYPE_INT_RGB);
	private Graphics grphOffBI = bfimgOffBI.getGraphics();
	private boolean blPaintBI = false;
	
	public static void main(String[] args) {
		TestKINECTV2_DepthAndBodyIndex testkinectv2 = new TestKINECTV2_DepthAndBodyIndex();
	}
	
	TestKINECTV2_DepthAndBodyIndex() {
		this.setTitle("TestKINECTV2_DepthAndBodyIndex");
		this.setSize(IMG_WIDTH * 2 + BORDER_WIDTH * 2, IMG_HEIGHT + TITLE_HEIGHT + PANEL_HEIGHT + BORDER_WIDTH);
		this.setLocationRelativeTo(null);
		this.setLayout(null);
		this.setResizable(false);
		
		imgcvs = new ImageCanvas();
		imgcvs.setBounds(0, 0, IMG_WIDTH, IMG_HEIGHT);
		getContentPane().add(imgcvs);
		
		imgcvsBI = new ImageCanvasBI();
		imgcvsBI.setBounds(IMG_WIDTH, 0, IMG_WIDTH, IMG_HEIGHT);
		getContentPane().add(imgcvsBI);
		
		pnl = new JPanel();
		pnl.setLayout(new FlowLayout(FlowLayout.RIGHT, 10, 5));
		pnl.setBounds(0, IMG_HEIGHT, IMG_WIDTH * 2, PANEL_HEIGHT);
		
		btnExit = new JButton("  Exit  ");
		btnExit.addActionListener(this);
		pnl.add(btnExit);
		
		getContentPane().add(pnl);
		
		this.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent ev) {
				System.exit(0);
			}
		});

		this.setVisible(true);
		
		Thread thread = new Thread(this);
		thread.start();
	}

	public void actionPerformed(ActionEvent e) {
		JButton btn = (JButton)e.getSource();
		
		if (btn == btnExit) {
			blExit = true;
		}
	}

	private class ImageCanvas extends Canvas {
		public void paint(Graphics g){
			if (blPaint == true) {
				g.drawImage(bfimgOff, 0, 0, this);
				blPaint = false;
			}
		}
		
		public void update(Graphics g){
			paint(g);
		}
	}

	private class ImageCanvasBI extends Canvas {
		public void paint(Graphics g){
			if (blPaintBI == true) {
				g.drawImage(bfimgOffBI, 0, 0, this);
				blPaintBI = false;
			}
		}
		
		public void update(Graphics g){
			paint(g);
		}
	}

	public void run() {
		try {
			initialize();
			drawcvs();
		}
		catch (Exception_KINECT e) {
			System.out.println("***** Exception_KINECT: " + e.getMessage());
		}
		finally {
			System.out.println("finally start ");
			try {
				terminate();
			}
			catch (Exception_KINECT e) {
				System.out.println("***** Exception_KINECT: " + e.getMessage());
			}
			System.out.println("finally end ");
		}
		System.exit(0);
	}

	void initialize() throws Exception_KINECT {
		// sensor
		wknct.GetDefaultKinectSensor(sensor);
		System.out.println("sensor: " + Long.toHexString(sensor[0]));
		
		wiks = new Wrp_IKinectSensor(sensor[0]);
		wiks.Open();
		
		// depth source
		wiks.get_DepthFrameSource(source);
		System.out.println("source: " + Long.toHexString(source[0]));
		
		widfs = new Wrp_IDepthFrameSource(source[0]);
		
		// depth reader
		widfs.OpenReader(reader);
		System.out.println("reader: " + Long.toHexString(reader[0]));
		
		widfr = new Wrp_IDepthFrameReader(reader[0]);
		
		// depth description
		widfs.get_FrameDescription(descript);
		System.out.println("descript: " + Long.toHexString(descript[0]));
		
		wifd = new Wrp_IFrameDescription(descript[0]);
		
		int[] pWidth = {0};
		int[] pHeight = {0};
		wifd.get_Width(pWidth);
		wifd.get_Height(pHeight);
		System.out.println("width: " + Integer.toString(pWidth[0])
				+ " height: " + Integer.toString(pHeight[0]));
		width = pWidth[0];
		height = pHeight[0];
		
		color = new int[width * height];
		
		// depth reliable distance
		short[] pDistanceMin = {0};
		short[] pDistanceMax = {0};
		widfs.get_DepthMinReliableDistance(pDistanceMin);
		widfs.get_DepthMaxReliableDistance(pDistanceMax);
		System.out.println("distanceMin: " + Short.toString(pDistanceMin[0])
				+ " distanceMax: " + Short.toString(pDistanceMax[0]));
		distanceMin = pDistanceMin[0];
		distanceMax = pDistanceMax[0];

		// body index source
		wiks.get_BodyIndexFrameSource(sourceBI);
		System.out.println("sourceBI: " + Long.toHexString(sourceBI[0]));
		
		wibifs = new Wrp_IBodyIndexFrameSource(sourceBI[0]);
		
		// body index reader
		wibifs.OpenReader(readerBI);
		System.out.println("readerBI: " + Long.toHexString(readerBI[0]));
		
		wibifr = new Wrp_IBodyIndexFrameReader(readerBI[0]);
		
		// body index description
		wibifs.get_FrameDescription(descriptBI);
		System.out.println("descriptBI: " + Long.toHexString(descriptBI[0]));
		
		wifdbi = new Wrp_IFrameDescription(descriptBI[0]);
		
		int[] pWidthBI = {0};
		int[] pHeightBI = {0};
		wifdbi.get_Width(pWidthBI);
		wifdbi.get_Height(pHeightBI);
		System.out.println("widthBI: " + Integer.toString(pWidthBI[0])
				+ " heightBI: " + Integer.toString(pHeightBI[0]));
		widthBI = pWidthBI[0];
		heightBI = pHeightBI[0];
		
		colorBI = new int[widthBI * heightBI];

	}
	
	void drawcvs() throws Exception_KINECT {
		
		for (;;) {
			
			if (setDepthImage()) {
				blPaint = true;
				imgcvs.repaint();
			}
			
			if (setBodyIndexImage()) {
				blPaintBI = true;
				imgcvsBI.repaint();
			}
			
			if (blExit) {
				break;
			}
		}
	}
	
	boolean setDepthImage() throws Exception_KINECT {
		boolean blSet = false;
		
		// depth frame
		frame[0] = 0;
		int hr = widfr.AcquireLatestFrame(frame);
		System.out.println("frame: " + Long.toHexString(frame[0]));
		
		if (H_WinError.SUCCEEDED(hr)) {
			widf = new Wrp_IDepthFrame(frame[0]);
			
			int[] capacity = {0};
			short[] buffer = widf.AccessUnderlyingBuffer(capacity);
			System.out.println("AccessUnderlyingBuffer capacity: " + Integer.toString(capacity[0])
					+ " buffer.length: " + Integer.toString(buffer.length));
			
			for (int i = 0; i < color.length; i++) {
				if (distanceMin <= buffer[i] && buffer[i] <= distanceMax) {
					int gray = 0xff - buffer[i] * 0xff / distanceMax;
					color[i] = (gray << 16) | (gray << 8) | gray; 
				}
				else if (buffer[i] == 0) {
					color[i] = 0x00ff0000; // red is non data
				}
				else {
					color[i] = 0x0000ff00; // green is range over
				}
			}
			
			// set color image
			bfimgOff.setRGB(0, 0, IMG_WIDTH, IMG_HEIGHT, color, 0, IMG_WIDTH);
			
			blSet = true;
		}
		
		if(widf != null) widf.Release();
		widf = null;
		frame[0] = 0;
		return blSet;
	}
	
	boolean setBodyIndexImage() throws Exception_KINECT {
		boolean blSet = false;

		// body index frame
		frameBI[0] = 0;
		int  hr = wibifr.AcquireLatestFrame(frameBI);
		System.out.println("frameBI: " + Long.toHexString(frameBI[0]));
		
		if (H_WinError.SUCCEEDED(hr)) {
			wibif = new Wrp_IBodyIndexFrame(frameBI[0]);
			
			int[] capacity = {0};
			byte[] buffer = wibif.AccessUnderlyingBuffer(capacity);
			System.out.println("AccessUnderlyingBuffer capacity: " + Integer.toString(capacity[0])
					+ " buffer.length: " + Integer.toString(buffer.length));

			for (int i = 0; i < colorBI.length; i++) {
				if (0 <= buffer[i] && buffer[i] <= 5) {
					colorBI[i] = colorTable[buffer[i]]; 
				}
				else {
					colorBI[i] = 0x00404040; // gray
				}
			}
			
			// set color image
			bfimgOffBI.setRGB(0, 0, IMG_WIDTH, IMG_HEIGHT, colorBI, 0, IMG_WIDTH);
			
			blSet = true;
		}
		
		if (wibif != null) wibif.Release();
		wibif = null;
		frameBI[0] = 0;
		return blSet;
	}
	
	void terminate() throws Exception_KINECT {
		if (widf != null) widf.Release();
		if (widfr != null) widfr.Release();
		if (wifd != null) wifd.Release();
		if (widfs != null) widfs.Release();
		if (wibif != null) wibif.Release();
		if (wibifr != null) wibifr.Release();
		if (wifdbi != null) wifdbi.Release();
		if (wibifs != null) wibifs.Release();
		
		if (wiks != null) {
			wiks.Close();
			wiks.Release();
		}
		
		if (grphOff != null) grphOff.dispose();
		if (grphOffBI != null) grphOffBI.dispose();
	}

}
