package testkinectv2;

import javax.swing.*;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.Collection;
import java.util.LinkedList;
import java.util.concurrent.*;

public class TestKINECTV2_EventAndMultiSource extends JFrame implements Runnable, ActionListener, ComponentListener {

	// Kinect color, depth and body index 
	private long[] sensor = {0};
	private long[] readerM = {0};
	private long[] argsM = {0};
	private long[] referenceM = {0};
	private long[] mapper = {0};
	private long[] frameM = {0};
	private long[] referenceC = {0};
	private long[] frameC = {0};
	private long[] referenceD = {0};
	private long[] frameD = {0};
	private long[] referenceBI = {0};
	private long[] frameBI = {0};
	private long[] hMulti = {H_WinBase.INVALID_HANDLE_VALUE};

	private int[] color = null;
	private int[] colorBack = null;
	private int[] colorMerge = null;

	private int width = 0, height = 0;
	private int widthBI = 0, heightBI = 0;
	private final String BACKGROUND_IMAGE_FILE = "backimgv2.jpg";
	private final int BACK_IMG_WIDTH = 1920;
	private final int BACK_IMG_HEIGHT = 1080;
	
	// thread pool
	private ExecutorService threadPool = Executors.newFixedThreadPool(8);
	private Collection<Callable<Void>> processes = new LinkedList<Callable<Void>>();
	private final int COLOR_INDEX_STEP = 64800;
	
	// wrapper
	private Wrp_KINECTV2 wknct = new Wrp_KINECTV2();
	private Wrp_Kernel32 wkrnl = new Wrp_Kernel32();
	private Wrp_IKinectSensor wiks = null;
	private Wrp_ICoordinateMapper wicm = null;
	private Wrp_IMultiSourceFrameReader wimsfr = null;
	private Wrp_IMultiSourceFrameArrivedEventArgs wimsfaea = null;
	private Wrp_IMultiSourceFrameReference wimsfref = null;
	private Wrp_IMultiSourceFrame wimsf = null;
	private Wrp_IColorFrameReference wicfref = null;
	private Wrp_IColorFrame wicf = null;
	private Wrp_IDepthFrameReference widfref = null;
	private Wrp_IDepthFrame widf = null;
	private Wrp_IBodyIndexFrameReference wibifref = null;
	private Wrp_IBodyIndexFrame wibif = null;
		
	// GUI
	private final int IMG_WIDTH = 1920 / 2;
	private final int IMG_HEIGHT = 1080 / 2;
	private final int BORDER_WIDTH = 4;
	private final int TITLE_HEIGHT = 30;
	private final int PANEL_HEIGHT = 40;
	private JButton btnExit;
	private boolean blExit = false; 
	private JPanel pnl;
	private ImageCanvas imgcvs;
	private BufferedImage bfimgOff = null;
	private Graphics2D grphOff = null;
	private boolean blPaint = false;
	
	
	public static void main(String[] args) {
		TestKINECTV2_EventAndMultiSource testkinectv2 = new TestKINECTV2_EventAndMultiSource();
	}
	
	TestKINECTV2_EventAndMultiSource() {
		this.setTitle("TestKINECTV2_EventAndMultiSource");
		this.setSize(IMG_WIDTH + BORDER_WIDTH * 2, IMG_HEIGHT + TITLE_HEIGHT + PANEL_HEIGHT + BORDER_WIDTH);
		this.addComponentListener(this);
		this.setLocationRelativeTo(null);
		this.setLayout(null);
		this.setResizable(true);
		getContentPane().setBackground(Color.DARK_GRAY);
		
		imgcvs = new ImageCanvas();
		imgcvs.setBounds(0, 0, IMG_WIDTH, IMG_HEIGHT);
		getContentPane().add(imgcvs);
		
		pnl = new JPanel();
		pnl.setLayout(new FlowLayout(FlowLayout.RIGHT, 10, 5));
		pnl.setBounds(0, IMG_HEIGHT, IMG_WIDTH, PANEL_HEIGHT);
		
		btnExit = new JButton("  Exit  ");
		btnExit.addActionListener(this);
		pnl.add(btnExit);
		
		getContentPane().add(pnl);
		
		this.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent ev) {
				System.exit(0);
			}
		});

		this.setVisible(true);
		
		Thread thread = new Thread(this);
		thread.start();
	}

	public void actionPerformed(ActionEvent e) {
		JButton btn = (JButton)e.getSource();
		
		if (btn == btnExit) {
			blExit = true;
		}
	}

	public void componentResized(ComponentEvent e) {
		Rectangle rec = e.getComponent().getBounds();
		rec.width -= BORDER_WIDTH * 2;
		rec.height -= TITLE_HEIGHT + PANEL_HEIGHT + BORDER_WIDTH;
		
		System.out.println("componentResized width: " + Integer.toString(rec.width)
				+ " height: " + Integer.toString(rec.height));

		if ((double)IMG_HEIGHT / (double)IMG_WIDTH >= (double)rec.height / (double)rec.width) {
			int w = (int)((double)rec.height * (double)IMG_WIDTH / (double)IMG_HEIGHT);
			imgcvs.setBounds((rec.width - w) / 2, 0, w, rec.height);
		}
		else {
			int h = (int)((double)rec.width * (double)IMG_HEIGHT / (double)IMG_WIDTH);
			imgcvs.setBounds(0, (rec.height - h) / 2, rec.width, h);
		}
		
		pnl.setBounds(0, rec.height, rec.width, PANEL_HEIGHT);
	}

	public void componentMoved(ComponentEvent e) {
	}

	public void componentShown(ComponentEvent e) {
	}

	public void componentHidden(ComponentEvent e) {
	}

	private class ImageCanvas extends Canvas {
		public void paint(Graphics g){
			if (blPaint == true) {
				g.drawImage(bfimgOff, 0, 0, imgcvs.getWidth(), imgcvs.getHeight(), this);
				
				blPaint = false;
			}
		}
		
		public void update(Graphics g){
			paint(g);
		}
	}

	public void run() {
		try {
			initialize();
			drawcvs();
		}
		catch (Exception_KINECT e) {
			System.out.println("***** Exception_KINECT: " + e.getMessage());
		}
		catch (Exception_Win64 e) {
			System.out.println("***** Exception_Win64: " + e.getMessage());
		}
		finally {
			System.out.println("finally start ");
			try {
				terminate();
			}
			catch (Exception_KINECT e) {
				System.out.println("***** Exception_KINECT: " + e.getMessage());
			}
			System.out.println("finally end ");
		}
		System.exit(0);
	}
	
	private void initialize() throws Exception_KINECT {
		// sensor
		wknct.GetDefaultKinectSensor(sensor);
		System.out.println("sensor: " + Long.toHexString(sensor[0]));
		
		wiks = new Wrp_IKinectSensor(sensor[0]);
		wiks.Open();
		
		// open multi reader
		final int types = H_Kinect.FrameSourceTypes_Color
				| H_Kinect.FrameSourceTypes_Depth
				| H_Kinect.FrameSourceTypes_BodyIndex;
		
		wiks.OpenMultiSourceFrameReader(types, readerM);
		System.out.println("readerM: " + Long.toHexString(readerM[0]));
		
		wimsfr = new Wrp_IMultiSourceFrameReader(readerM[0]);
		
		// subscribe multi arrived
		wimsfr.SubscribeMultiSourceFrameArrived(hMulti);
		System.out.println("hMulti: " + Long.toHexString(hMulti[0]));
		
		width = BACK_IMG_WIDTH;
		height = BACK_IMG_HEIGHT;
		
		color = new int[width * height];
		colorMerge = new int[width * height];
		
		widthBI = 512;
		heightBI = 424;
		
		// coordinate mapper
		wiks.get_CoordinateMapper(mapper);
		System.out.println("mapper: " + Long.toHexString(mapper[0]));
		
		wicm = new Wrp_ICoordinateMapper(mapper[0]);
		
		// read background image 
		colorBack = new int[BACK_IMG_WIDTH * BACK_IMG_HEIGHT];
		readBackImage();
	}
	
	private void readBackImage() {
		Image backImage = Toolkit.getDefaultToolkit().getImage(BACKGROUND_IMAGE_FILE);
        PixelGrabber pg = new PixelGrabber(backImage, 0, 0, BACK_IMG_WIDTH, BACK_IMG_HEIGHT, colorBack, 0, BACK_IMG_WIDTH);
        try {
           pg.grabPixels();
        } catch(InterruptedException ie) {}
	}

	private void drawcvs() throws Exception_KINECT, Exception_Win64 {
		bfimgOff = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
		grphOff = (Graphics2D)bfimgOff.getGraphics();
		
		for (;;) {
			
			if (setColorImage()) {
				blPaint = true;
				imgcvs.repaint();
			}
			
			if (blExit) {
				break;
			}
		}
	}

	private boolean setColorImage() throws Exception_KINECT, Exception_Win64 {
		boolean blSet = false;
		
		// wait single event
		int objectNo = wkrnl.WaitForSingleObject(hMulti[0], H_WinBase.INFINITE);
		System.out.println("objectNo: " + Integer.toString(objectNo));
		
		if (objectNo != H_WinBase.WAIT_OBJECT_0) return blSet;
		
		// get event data
		argsM[0] = 0;
		wimsfr.GetMultiSourceFrameArrivedEventData(hMulti[0], argsM);
		System.out.println("argsM: " + Long.toHexString(argsM[0]));
	
		wimsfaea = new Wrp_IMultiSourceFrameArrivedEventArgs(argsM[0]);
		
		// get reference
		referenceM[0] = 0;
		wimsfaea.get_FrameReference(referenceM);
		System.out.println("referenceM: " + Long.toHexString(referenceM[0]));
		
		wimsfref = new Wrp_IMultiSourceFrameReference(referenceM[0]);
		
		// acquire frame
		frameM[0] = 0;
		wimsfref.AcquireFrame(frameM);
		System.out.println("frameM: " + Long.toHexString(frameM[0]));
		
		wimsf = new Wrp_IMultiSourceFrame(frameM[0]);
		
		// get color reference
		referenceC[0] = 0;
		wimsf.get_ColorFrameReference(referenceC);
		System.out.println("referenceC: " + Long.toHexString(referenceC[0]));
		
		wicfref = new Wrp_IColorFrameReference(referenceC[0]);
		
		// get color frame
		frameC[0] = 0;
		int hr = wicfref.AcquireFrame(frameC);
		System.out.println("frameC: " + Long.toHexString(frameC[0]));
		
		if (H_WinError.SUCCEEDED(hr) && frameC[0] != 0) {
			wicf = new Wrp_IColorFrame(frameC[0]);
			
			int capacity = color.length * 4;
			wicf.CopyConvertedFrameDataToArray(capacity, color, H_Kinect.ColorImageFormat_Bgra);
			
			// get depth data
			short[] depthData = getDepthData();
			
			// get body index data
			byte[] bodyIndexData = getBodyIndexData();
		
			if (depthData != null && bodyIndexData != null) {
				// get depth points for color frame
				float[] depthPoints = new float[color.length * 2]; // DepthSpacePoint {float X, float Y;}
				wicm.MapColorFrameToDepthSpace(depthData.length, depthData, color.length, depthPoints);
				
				for (int ii = 0; ii < color.length; ii += COLOR_INDEX_STEP) {
					final int start = ii;
					final int end = start + COLOR_INDEX_STEP;
					final float[] depthPointsFinal = depthPoints;
					final byte[] bodyIndexDataFinal = bodyIndexData;
					
					processes.add(new Callable<Void>() {
				        @Override
				        public Void call() throws Exception_KINECT {
					
							for (int i = start; i < end; i++) {
								if (chkFloat(depthPointsFinal[i * 2]) && chkFloat(depthPointsFinal[i * 2 + 1])) {
									int pointX = (int)depthPointsFinal[i * 2];
									int pointY = (int)depthPointsFinal[i * 2 + 1];
									
									if (chkRange(pointX, pointY)) {
										int index = pointX + pointY * widthBI;
										if (0 <= bodyIndexDataFinal[index] && bodyIndexDataFinal[index] <= 5) {
											colorMerge[i] = color[i];
											continue;
										}
									}
								}
								colorMerge[i] = colorBack[i];
							}
				            return null;
				        }
				    });
							
				}
				
				try {
				    threadPool.invokeAll(processes);
				} catch (InterruptedException e) {
				    throw new RuntimeException(e);
				} finally {
					processes.clear();
				}

				// set color image
				bfimgOff.setRGB(0, 0, width, height, colorMerge, 0, width);

				blSet = true;
			}
		}
		
		// release
		if(wicf != null) wicf.Release();
		wicf = null;
		frameC[0] = 0;
		if(wicfref != null) wicfref.Release();
		wicfref = null;
		referenceC[0] = 0;
		if(wimsf != null) wimsf.Release();
		wimsf = null;
		frameM[0] = 0;
		if(wimsfref != null) wimsfref.Release();
		wimsfref = null;
		referenceM[0] = 0;
		if(wimsfaea != null) wimsfaea.Release();
		wimsfaea = null;
		argsM[0] = 0;
		
		return blSet;
	}
	
	private short[] getDepthData() throws Exception_KINECT {
		short[] depthData = null;
		
		// get depth reference
		referenceD[0] = 0;
		wimsf.get_DepthFrameReference(referenceD);
		System.out.println("referenceD: " + Long.toHexString(referenceD[0]));
		
		widfref = new Wrp_IDepthFrameReference(referenceD[0]);
		 		
		// get depth frame
		frameD[0] = 0;
		int hr = widfref.AcquireFrame(frameD);
		System.out.println("frameD: " + Long.toHexString(frameD[0]));
		
		if (H_WinError.SUCCEEDED(hr) && frameD[0] != 0) {
			widf = new Wrp_IDepthFrame(frameD[0]);
			
			int[] capacity = {0};
			depthData = widf.AccessUnderlyingBuffer(capacity);
			System.out.println("AccessUnderlyingBuffer capacity: " + Integer.toString(capacity[0])
					+ " buffer.length: " + Integer.toString(depthData.length));
		}
		
		if(widf != null) widf.Release();
		widf = null;
		frameD[0] = 0;
		if(widfref != null) widfref.Release();
		widfref = null;
		referenceD[0] = 0;
		
		return depthData;
	}
	
	private byte[] getBodyIndexData() throws Exception_KINECT {
		byte[] bodyIndexData = null;

		// get body index reference
		referenceBI[0] = 0;
		wimsf.get_BodyIndexFrameReference(referenceBI);
		System.out.println("referenceBI: " + Long.toHexString(referenceBI[0]));
		
		wibifref = new Wrp_IBodyIndexFrameReference(referenceBI[0]);
		  		
		// get body index frame
		frameBI[0] = 0;
		int hr = wibifref.AcquireFrame(frameBI);
		System.out.println("frameBI: " + Long.toHexString(frameBI[0]));

		if (H_WinError.SUCCEEDED(hr) && frameBI[0] != 0) {
			wibif = new Wrp_IBodyIndexFrame(frameBI[0]);
				
			int[] capacity = {0};
			bodyIndexData = wibif.AccessUnderlyingBuffer(capacity);
			System.out.println("AccessUnderlyingBuffer capacity: " + Integer.toString(capacity[0])
					+ " buffer.length: " + Integer.toString(bodyIndexData.length));
		}
		
		if(wibif != null) wibif.Release();
		wibif = null;
		frameBI[0] = 0;
		if(wibifref != null) wibifref.Release();
		wibifref = null;
		referenceBI[0] = 0;

		return bodyIndexData;
	}

	private boolean chkFloat(float f) {
		if (Float.isNaN(f)) return false;
		if (Float.isInfinite(f)) return false;
		return true;
	}
	
	private boolean chkRange(int pointX, int pointY) {
		if (pointX < 0 || widthBI <= pointX) return false;
		if (pointY < 0 || heightBI <= pointY) return false;
		return true;
	}

	void terminate() throws Exception_KINECT {
		if (wicf != null) wicf.Release();
		if (wicfref != null) wicfref.Release();
		if (widf != null) widf.Release();
		if (widfref != null) widfref.Release();
		if (wibif != null) wibif.Release();
		if (wibifref != null) wibifref.Release();
		if (wimsf != null) wimsf.Release();
		if (wimsfref != null) wimsfref.Release();
		if (wimsfaea != null) wimsfaea.Release();
		if (wimsfr != null) wimsfr.Release();
		if (wicm != null) wicm.Release();
		
		if (wiks != null) {
			wiks.Close();
			wiks.Release();
		}
		
		if (grphOff != null) grphOff.dispose();
	}
	
}
