package testkinectv2;

import javax.swing.*;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.Collection;
import java.util.LinkedList;
import java.util.concurrent.*;

public class TestKINECTV2_Face extends JFrame implements Runnable, ActionListener, ComponentListener {
	
	// Kinect color, body and face
	private long[] sensor = {0};
	private long[] source = {0};
	private long[] reader = {0};
	private long[] descript = {0};
	private long[] mapper = {0};
	private long[] frame = {0};
	private long[] sourceB = {0};
	private long[] readerB = {0};
	private long[] descriptB = {0};
	private long[] frameB = {0};
	private long[] bodies = new long[H_Kinect.BODY_COUNT];
	private long[][] sourceF = new long[H_Kinect.BODY_COUNT][1];
	private long[][] readerF = new long[H_Kinect.BODY_COUNT][1];
	private long[][] frameF = new long[H_Kinect.BODY_COUNT][1];
	private long[][] resultF = new long[H_Kinect.BODY_COUNT][1];
	
	private int[] color = null;
	private int width = 0, height = 0;
	private String[] strFace = new String[H_Kinect.BODY_COUNT];
	private long[] trackingIdSave = new long[H_Kinect.BODY_COUNT];
	
	private final int features = H_Kinect_Face.FaceFrameFeatures_BoundingBoxInColorSpace
			| H_Kinect_Face.FaceFrameFeatures_PointsInColorSpace
			| H_Kinect_Face.FaceFrameFeatures_RotationOrientation
			| H_Kinect_Face.FaceFrameFeatures_Happy
			| H_Kinect_Face.FaceFrameFeatures_RightEyeClosed
			| H_Kinect_Face.FaceFrameFeatures_LeftEyeClosed
			| H_Kinect_Face.FaceFrameFeatures_MouthOpen
			| H_Kinect_Face.FaceFrameFeatures_MouthMoved
			| H_Kinect_Face.FaceFrameFeatures_LookingAway
			| H_Kinect_Face.FaceFrameFeatures_Glasses
			| H_Kinect_Face.FaceFrameFeatures_FaceEngagement;
	
	// thread pool
	private ExecutorService threadPool = Executors.newFixedThreadPool(H_Kinect.BODY_COUNT);
	private Collection<Callable<Void>> processes = new LinkedList<Callable<Void>>();

	// wrapper
	private Wrp_KINECTV2 wknct = new Wrp_KINECTV2();
	private Wrp_IKinectSensor wiks = null;
	private Wrp_IColorFrameSource wicfs = null;
	private Wrp_IColorFrameReader wicfr = null;
	private Wrp_IFrameDescription wifd = null;
	private Wrp_ICoordinateMapper wicm = null;
	private Wrp_IColorFrame wicf = null;
	private Wrp_IBodyFrameSource wibfs = null;
	private Wrp_IBodyFrameReader wibfr = null;
	private Wrp_IBodyFrame wibf = null;
	private Wrp_IBody[] wib = new Wrp_IBody[H_Kinect.BODY_COUNT];
	private Wrp_KINECTV2FACE wknctf = new Wrp_KINECTV2FACE();
	private Wrp_IFaceFrameSource[] wiffs = new Wrp_IFaceFrameSource[H_Kinect.BODY_COUNT];
	private Wrp_IFaceFrameReader[] wiffr = new Wrp_IFaceFrameReader[H_Kinect.BODY_COUNT];
	private Wrp_IFaceFrame[] wiff = new Wrp_IFaceFrame[H_Kinect.BODY_COUNT];
	private Wrp_IFaceFrameResult[] wiffrslt = new Wrp_IFaceFrameResult[H_Kinect.BODY_COUNT];
	
	// GUI
	private final int IMG_WIDTH = 1920 / 2;
	private final int IMG_HEIGHT = 1080 / 2;
	private final int BORDER_WIDTH = 4;
	private final int TITLE_HEIGHT = 30;
	private final int PANEL_HEIGHT = 40;
	private final int PANEL_BODY_HEIGHT = 120;
	private JCheckBox chkBox, chkProperty, chkRotation, chkBody;
	private JButton btnExit;
	private boolean blExit = false; 
	private JPanel pnl;
	private JLabel[] lblFace = new JLabel[H_Kinect.BODY_COUNT];
	private JPanel pnlFace;
	private ImageCanvas imgcvs;
	private BufferedImage bfimgOff = null;
	private Graphics2D grphOff = null;
	private boolean blPaint = false;
	private final Color[] colorJoint = {Color.RED, Color.GREEN, Color.BLUE, Color.YELLOW, Color.MAGENTA, Color.CYAN};
	private final Color[] colorText = {new Color(128, 64, 64), new Color(64, 128, 64), new Color(64, 64, 128), 
			new Color(128, 128, 64), new Color(128, 64, 128), new Color(64, 128, 128)};
	private final float LINE_WIDTH = 5.0f;

	public static void main(String[] args) {
		TestKINECTV2_Face testkinectv2 = new TestKINECTV2_Face();
	}
	
	TestKINECTV2_Face() {
		this.setTitle("TestKINECTV2_Face");
		this.setSize(IMG_WIDTH + BORDER_WIDTH * 2, IMG_HEIGHT + TITLE_HEIGHT + PANEL_BODY_HEIGHT + PANEL_HEIGHT + BORDER_WIDTH);
		this.addComponentListener(this);
		this.setLocationRelativeTo(null);
		this.setLayout(null);
		this.setResizable(true);
		getContentPane().setBackground(Color.DARK_GRAY);
		
		imgcvs = new ImageCanvas();
		imgcvs.setBounds(0, 0, IMG_WIDTH, IMG_HEIGHT);
		getContentPane().add(imgcvs);
		
		pnlFace = new JPanel();
		pnlFace.setLayout(new GridLayout(H_Kinect.BODY_COUNT, 1));
		pnlFace.setBounds(0, IMG_HEIGHT, IMG_WIDTH, PANEL_BODY_HEIGHT);
		
		for (int i = 0; i < lblFace.length; i++) {
			lblFace[i] = new JLabel();
			lblFace[i].setForeground(colorText[i]);
			lblFace[i].setFont(new Font(Font.MONOSPACED, Font.PLAIN , 12));
			pnlFace.add(lblFace[i]);
		}
		
		getContentPane().add(pnlFace);

		pnl = new JPanel();
		pnl.setLayout(new FlowLayout(FlowLayout.RIGHT, 10, 5));
		pnl.setBounds(0, IMG_HEIGHT + PANEL_BODY_HEIGHT, IMG_WIDTH, PANEL_HEIGHT);
		
		chkBox = new JCheckBox("BoundingBox", true);
		pnl.add(chkBox);
		
		chkProperty = new JCheckBox("Property", true);
		pnl.add(chkProperty);
		
		chkRotation = new JCheckBox("Rotation", true);
		pnl.add(chkRotation);
		
		chkBody = new JCheckBox("Body", true);
		pnl.add(chkBody);
		
		btnExit = new JButton("  Exit  ");
		btnExit.addActionListener(this);
		pnl.add(btnExit);
		
		getContentPane().add(pnl);
		
		this.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent ev) {
				System.exit(0);
			}
		});

		this.setVisible(true);
		
		Thread thread = new Thread(this);
		thread.start();
	}
	
	public void actionPerformed(ActionEvent e) {
		JButton btn = (JButton)e.getSource();
		
		if (btn == btnExit) {
			blExit = true;
		}
	}

	public void componentResized(ComponentEvent e) {
		Rectangle rec = e.getComponent().getBounds();
		rec.width -= BORDER_WIDTH * 2;
		rec.height -= TITLE_HEIGHT + PANEL_BODY_HEIGHT + PANEL_HEIGHT + BORDER_WIDTH;
		
		System.out.println("componentResized width: " + Integer.toString(rec.width)
				+ " height: " + Integer.toString(rec.height));

		if ((double)IMG_HEIGHT / (double)IMG_WIDTH >= (double)rec.height / (double)rec.width) {
			int w = (int)((double)rec.height * (double)IMG_WIDTH / (double)IMG_HEIGHT);
			imgcvs.setBounds((rec.width - w) / 2, 0, w, rec.height);
		}
		else {
			int h = (int)((double)rec.width * (double)IMG_HEIGHT / (double)IMG_WIDTH);
			imgcvs.setBounds(0, (rec.height - h) / 2, rec.width, h);
		}
		
		pnlFace.setBounds(0, rec.height, rec.width, PANEL_BODY_HEIGHT);
		pnl.setBounds(0, rec.height + PANEL_BODY_HEIGHT, rec.width, PANEL_HEIGHT);
	}

	public void componentMoved(ComponentEvent e) {
	}

	public void componentShown(ComponentEvent e) {
	}

	public void componentHidden(ComponentEvent e) {
	}

	private class ImageCanvas extends Canvas {
		public void paint(Graphics g){
			if (blPaint == true) {
				g.drawImage(bfimgOff, 0, 0, imgcvs.getWidth(), imgcvs.getHeight(), this);
				
				blPaint = false;
			}
		}
		
		public void update(Graphics g){
			paint(g);
		}
	}

	public void run() {
		try {
			initialize();
			drawcvs();
		}
		catch (Exception_KINECT e) {
			System.out.println("***** Exception_KINECT: " + e.getMessage());
		}
		finally {
			System.out.println("finally start ");
			try {
				terminate();
			}
			catch (Exception_KINECT e) {
				System.out.println("***** Exception_KINECT: " + e.getMessage());
			}
			System.out.println("finally end ");
		}
		System.exit(0);
	}

	private void initialize() throws Exception_KINECT {
		// sensor
		wknct.GetDefaultKinectSensor(sensor);
		System.out.println("sensor: " + Long.toHexString(sensor[0]));
		
		wiks = new Wrp_IKinectSensor(sensor[0]);
		wiks.Open();
		
		// color source
		wiks.get_ColorFrameSource(source);
		System.out.println("source: " + Long.toHexString(source[0]));
		
		wicfs = new Wrp_IColorFrameSource(source[0]);
		
		// color reader
		wicfs.OpenReader(reader);
		System.out.println("reader: " + Long.toHexString(reader[0]));
		
		wicfr = new Wrp_IColorFrameReader(reader[0]);
		
		// color description
		wicfs.get_FrameDescription(descript);
		System.out.println("descript: " + Long.toHexString(descript[0]));
		
		wifd = new Wrp_IFrameDescription(descript[0]);
		
		int[] pWidth = {0};
		int[] pHeight = {0};
		wifd.get_Width(pWidth);
		wifd.get_Height(pHeight);
		System.out.println("width: " + Integer.toString(pWidth[0])
				+ " height: " + Integer.toString(pHeight[0]));
		width = pWidth[0];
		height = pHeight[0];
		
		color = new int[width * height];
		
		// body source
		wiks.get_BodyFrameSource(sourceB);
		System.out.println("sourceB: " + Long.toHexString(sourceB[0]));
		
		wibfs = new Wrp_IBodyFrameSource(sourceB[0]);
		
		// body reader
		wibfs.OpenReader(readerB);
		System.out.println("readerB: " + Long.toHexString(readerB[0]));
		
		wibfr = new Wrp_IBodyFrameReader(readerB[0]);
		
		// coordinate mapper
		wiks.get_CoordinateMapper(mapper);
		System.out.println("mapper: " + Long.toHexString(mapper[0]));
		
		wicm = new Wrp_ICoordinateMapper(mapper[0]);
		
		for (int i = 0; i < H_Kinect.BODY_COUNT; i++) {
			// face source
			sourceF[i][0] = 0;
			wknctf.CreateFaceFrameSource(sensor[0], 0, features, sourceF[i]);
			System.out.println("sourceF#" + Integer.toString(i) + ": " + Long.toHexString(sourceF[i][0]));
			
			wiffs[i] = new Wrp_IFaceFrameSource(sourceF[i][0]);
			
			// face reader
			readerF[i][0] = 0;
			wiffs[i].OpenReader(readerF[i]);
			System.out.println("readerF#" + Integer.toString(i) + ": " + Long.toHexString(readerF[i][0]));
			
			wiffr[i] = new Wrp_IFaceFrameReader(readerF[i][0]);
		}
		
		// initialize array
		for (int i = 0; i < H_Kinect.BODY_COUNT; i++) {
			bodies[i] = 0;
			wib[i] = null;
			
			frameF[i][0] = 0;
			wiff[i] = null;
			
			resultF[i][0] = 0;
			wiffrslt[i] = null;
		}
		
	}
	
	private void drawcvs() throws Exception_KINECT {
		bfimgOff = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
		grphOff = (Graphics2D)bfimgOff.getGraphics();
		
		for (;;) {
			
			if (setColorImage()) {
				blPaint = true;
				imgcvs.repaint();
			}
			
			if (blExit) {
				break;
			}
		}
	}

	private boolean setColorImage() throws Exception_KINECT {
		boolean blSet = false;
		
		// color frame
		frame[0] = 0;
		int hr = wicfr.AcquireLatestFrame(frame);
		System.out.println("frame: " + Long.toHexString(frame[0]));
		
		if (H_WinError.SUCCEEDED(hr)) {
			wicf = new Wrp_IColorFrame(frame[0]);
			
			int capacity = color.length * 4;
			wicf.CopyConvertedFrameDataToArray(capacity, color, H_Kinect.ColorImageFormat_Bgra);
			
			// set color image
			bfimgOff.setRGB(0, 0, width, height, color, 0, width);
			
			// draw bodies
			drawBody();

			blSet = true;
		}
		
		if (wicf != null) wicf.Release();
		wicf = null;
		frame[0] = 0;
		return blSet;
	}
	
	private boolean drawBody() throws Exception_KINECT {
		boolean blSet = false;

		// body frame 
		frameB[0] = 0;
		int  hr = wibfr.AcquireLatestFrame(frameB);
		System.out.println("frameB: " + Long.toHexString(frameB[0]));
		
		if (H_WinError.SUCCEEDED(hr) && frameB[0] != 0) {
			wibf = new Wrp_IBodyFrame(frameB[0]);
			
			for (int i = 0; i < bodies.length; i++) {
				bodies[i] = 0;
			}
			wibf.GetAndRefreshBodyData(bodies.length, bodies);
			System.out.println("bodies: " + Long.toHexString(bodies[0])
					+ " " + Long.toHexString(bodies[1]) + " " + Long.toHexString(bodies[2])
					+ " " + Long.toHexString(bodies[3]) + " " + Long.toHexString(bodies[4])
					+ " " + Long.toHexString(bodies[5]));
			
			for (int ii = 0; ii < bodies.length; ii++) {
				final int i = ii;
				processes.add(new Callable<Void>() {
			        @Override
			        public Void call() throws Exception_KINECT {
						strFace[i] = null;
						wib[i] = null;
						
						if (bodies[i] != 0) {
							wib[i] = new Wrp_IBody(bodies[i]);
							
							byte[] tracked = {0};
							wib[i].get_IsTracked(tracked);
							System.out.println("tracked: " + Byte.toString(tracked[0]));
							
							if (tracked[0] != 0) {
								
								if (chkBody.isSelected()) {
									Struct_Joint[] joints = new Struct_Joint[H_Kinect.JointType_Count];
									for (int j = 0; j < joints.length; j++) {
										joints[j] = new Struct_Joint();
									}
									wib[i].GetJoints(joints.length, joints);
									
									Struct_ColorSpacePoint[] colorPoints = new Struct_ColorSpacePoint[joints.length];
									for (int j = 0; j < colorPoints.length; j++) {
										if (joints[j].TrackingState != H_Kinect.TrackingState_NotTracked) {
											if (chkFloat(joints[j].Position.X) && chkFloat(joints[j].Position.Y) && chkFloat(joints[j].Position.Z)) {
												colorPoints[j] = new Struct_ColorSpacePoint();
												wicm.MapCameraPointToColorSpace(joints[j].Position, colorPoints[j]);
											}
										}
									}
								
									// draw joints
									drawJoints(colorPoints, i);
								}
								
								// set tracking ID
								long[] trackingId = {0};
								wib[i].get_TrackingId(trackingId);
								System.out.println("trackingId: " + Long.toHexString(trackingId[0]));
								if (trackingId[0] != trackingIdSave[i]) {
									wiffs[i].put_TrackingId(trackingId[0]);
									trackingIdSave[i] = trackingId[0];
								}
								
								// draw faces
								drawFace(i);
							}
						}
						if (strFace[i] == null) strFace[i] = "Face#" + Integer.toString(i) + " NotTracked";
						lblFace[i].setText(strFace[i]);
			            return null;
			        }
			    });
						
			}
			
			try {
			    threadPool.invokeAll(processes);
			} catch (InterruptedException e) {
			    throw new RuntimeException(e);
			} finally {
				processes.clear();
			}

			for (int i = 0; i < wib.length; i++) {
				if (wib[i] != null) wib[i].Release();
				wib[i] = null;
				bodies[i] = 0;
			}
			
			blSet = true;
		}
		
		if (wibf != null) wibf.Release();
		wibf = null;
		frameB[0] = 0;
		return blSet;
	}

	private boolean drawFace(int bodyNo) throws Exception_KINECT {
		boolean blSet = false;
	
		frameF[bodyNo][0] = 0;
		int hr = wiffr[bodyNo].AcquireLatestFrame(frameF[bodyNo]);
		System.out.println("frameF: " + Long.toHexString(frameF[bodyNo][0]));
		
		if (H_WinError.SUCCEEDED(hr) && frameF[bodyNo][0] != 0) {
		
			wiff[bodyNo] = new Wrp_IFaceFrame(frameF[bodyNo][0]);
			byte[] isTrackingIdValid = {0};
			wiff[bodyNo].get_IsTrackingIdValid(isTrackingIdValid);
			
			if (isTrackingIdValid[0] != 0) {
				resultF[bodyNo][0] = 0;
				wiff[bodyNo].get_FaceFrameResult(resultF[bodyNo]);
				
				wiffrslt[bodyNo] = new Wrp_IFaceFrameResult(resultF[bodyNo][0]);
				
				Struct_PointF[] points = new Struct_PointF[H_Kinect_Face.FacePointType_Count];
				for (int i = 0; i < points.length; i++) {
					points[i] = new Struct_PointF();
				}
				
				// get face points
				wiffrslt[bodyNo].GetFacePointsInColorSpace(points.length, points);
				
				// draw face points
				drawFacePoints(points, bodyNo);
				
				// draw face bounding box
				if (chkBox.isSelected()) {
					Struct_RectI rect = new Struct_RectI();
					wiffrslt[bodyNo].get_FaceBoundingBoxInColorSpace(rect);
					
					// draw box
					drawBox(rect, bodyNo);
				}
				
				// get face property
				strFace[bodyNo] = "Face#" + Integer.toString(bodyNo) + " Tracked   ";
				if(chkProperty.isSelected()) {
					int[] properties = new int[H_Kinect_Face.FaceProperty_Count];
					wiffrslt[bodyNo].GetFaceProperties(properties.length, properties);
					strFace[bodyNo] += toStrProperty(properties);
				}
				
				// get face rotation
				if(chkRotation.isSelected()) {
					Struct_Vector4 vec4 = new Struct_Vector4();
					wiffrslt[bodyNo].get_FaceRotationQuaternion(vec4);
					double[] pitch = {0.0d}, yaw = {0.0d}, roll = {0.0d};
					ExtractFaceRotationInDegrees(vec4, pitch, yaw, roll);
					strFace[bodyNo] += String.format("Rotation Pitch:%10.3f", pitch[0])
							+ String.format(" Yaw:%10.3f", yaw[0])
							+ String.format(" Roll:%10.3f", roll[0]);
				}
				
				if (wiffrslt[bodyNo] != null) wiffrslt[bodyNo].Release();
				wiffrslt[bodyNo] = null;
				resultF[bodyNo][0] = 0;
				blSet = true;
			}
		}
		
		if (wiff[bodyNo] != null) wiff[bodyNo].Release();
		wiff[bodyNo] = null;
		frameF[bodyNo][0] = 0;
		return blSet;
	}
	
	private void drawFacePoints(Struct_PointF[] points, int bodyNo) {
		grphOff.setStroke(new BasicStroke(LINE_WIDTH));
		grphOff.setColor(colorJoint[bodyNo]);
		final int r = 5;
		final int r2 = r * 2;
		
		for (int i = 0; i < points.length; i++) {
			if (chkFloat(points[i].X) && chkFloat(points[i].Y)) {
				int pointX = (int)points[i].X;
				int pointY = (int)points[i].Y;
				
				if (chkRange(pointX, pointY)) {
					grphOff.fillOval(pointX - r, pointY - r, r2, r2);
				}
			}
		}
	}

	private void drawBox(Struct_RectI rect, int bodyNo) {
		grphOff.setStroke(new BasicStroke(LINE_WIDTH));
		grphOff.setColor(colorJoint[bodyNo]);
		if (chkRange(rect.Left, rect.Top) && chkRange(rect.Right, rect.Bottom)) {
			grphOff.drawRect(rect.Left, rect.Top, rect.Right - rect.Left + 1, rect.Bottom - rect.Top + 1);
		}
	}
	
	private void ExtractFaceRotationInDegrees(Struct_Vector4 vec4Quaternion, double[] pPitch, double[] pYaw, double[] pRoll) {
        double x = vec4Quaternion.x;
        double y = vec4Quaternion.y;
        double z = vec4Quaternion.z;
        double w = vec4Quaternion.w;
		
        pPitch[0] = Math.atan2(2 * ((y * z) + (w * x)), (w * w) - (x * x) - (y * y) + (z * z)) / Math.PI * 180.0;
        pYaw[0] = Math.asin(2 * ((w * y) - (x * z))) / Math.PI * 180.0;
        pRoll[0] = Math.atan2(2 * ((x * y) + (w * z)), (w * w) + (x * x) - (y * y) - (z * z)) / Math.PI * 180.0;
	}
	
	private String toStrProperty(int[] propertyDetection) {
		String str = " Happy:" + toStrResult(propertyDetection[H_Kinect_Face.FaceProperty_Happy])
				+ " Engaged:" + toStrResult(propertyDetection[H_Kinect_Face.FaceProperty_Engaged])
				+ " WearingGlasses:" + toStrResult(propertyDetection[H_Kinect_Face.FaceProperty_WearingGlasses])
				+ " LeftEyeClosed:" + toStrResult(propertyDetection[H_Kinect_Face.FaceProperty_LeftEyeClosed])
				+ " RightEyeClosed:" + toStrResult(propertyDetection[H_Kinect_Face.FaceProperty_RightEyeClosed])
				+ " MouthOpen:" + toStrResult(propertyDetection[H_Kinect_Face.FaceProperty_MouthOpen])
				+ " MouthMoved:" + toStrResult(propertyDetection[H_Kinect_Face.FaceProperty_MouthMoved])
				+ " LookingAway:" + toStrResult(propertyDetection[H_Kinect_Face.FaceProperty_LookingAway]);
		return str;
	}
	
	private String toStrResult(int result) {
		String[] strResult = {"Unknown", "No     ", "Maybe  ", "Yes    "}; 
		String str = "";

		if (0 <= result && result < strResult.length) {
	    	str = strResult[result];
	    }
		
		return str;
	}
	
	private boolean chkFloat(float f) {
		if (Float.isNaN(f)) return false;
		if (Float.isInfinite(f)) return false;
		return true;
	}
	
	private boolean chkRange(int pointX, int pointY) {
		if (pointX < 0 || width <= pointX) return false;
		if (pointY < 0 || height <= pointY) return false;
		return true;
	}
	
	private boolean setPolyline(Struct_ColorSpacePoint[] colorPoints, int[] pointX, int[] pointY) {
		
		for (int i = 0; i < colorPoints.length; i++) {
			if (colorPoints[i] == null) return false;
			if (!chkFloat(colorPoints[i].X) || !chkFloat(colorPoints[i].Y)) return false;
			pointX[i] = (int)colorPoints[i].X;
			pointY[i] = (int)colorPoints[i].Y;
			if (!chkRange(pointX[i], pointY[i])) return false;
		}
	
		return true;
	}
	
	private void drawJoints(Struct_ColorSpacePoint[] colorPoints, int bodyNo) {
		
		grphOff.setStroke(new BasicStroke(LINE_WIDTH));
		grphOff.setColor(colorJoint[bodyNo]);
		final int r = 5;
		final int r2 = r * 2;
		
		// draw joints
		for (int i = 0; i < colorPoints.length; i++) {
			if (colorPoints[i] != null) {
				if (chkFloat(colorPoints[i].X) && chkFloat(colorPoints[i].Y)) {
					int pointX = (int)colorPoints[i].X;
					int pointY = (int)colorPoints[i].Y;
					
					if (chkRange(pointX, pointY)) {
						grphOff.drawOval(pointX - r, pointY - r, r2, r2);
					}
				}
			}
		}
		
		// draw lines
		int[] pointx = new int[6];
		int[] pointy = new int[6];
		
		grphOff.setStroke(new BasicStroke(LINE_WIDTH));
		grphOff.setColor(colorJoint[bodyNo]);

		
		final Struct_ColorSpacePoint[] headSpine = {
				colorPoints[H_Kinect.JointType_Head],
				colorPoints[H_Kinect.JointType_Neck],
				colorPoints[H_Kinect.JointType_SpineShoulder],
				colorPoints[H_Kinect.JointType_SpineMid],
				colorPoints[H_Kinect.JointType_SpineBase]};
		
		if (setPolyline(headSpine, pointx, pointy)) {
			grphOff.drawPolyline(pointx, pointy, headSpine.length);
		}
		
		final Struct_ColorSpacePoint[] handRight = {
				colorPoints[H_Kinect.JointType_SpineShoulder],
				colorPoints[H_Kinect.JointType_ShoulderRight],
				colorPoints[H_Kinect.JointType_ElbowRight],
				colorPoints[H_Kinect.JointType_WristRight],
				colorPoints[H_Kinect.JointType_HandRight],
				colorPoints[H_Kinect.JointType_HandTipRight]};
		
		if (setPolyline(handRight, pointx, pointy)) {
			grphOff.drawPolyline(pointx, pointy, handRight.length);
		}
		
		final Struct_ColorSpacePoint[] thumbRight = {
				colorPoints[H_Kinect.JointType_HandRight],
				colorPoints[H_Kinect.JointType_ThumbRight]};
		
		if (setPolyline(thumbRight, pointx, pointy)) {
			grphOff.drawPolyline(pointx, pointy, thumbRight.length);
		}
		
		final Struct_ColorSpacePoint[] handLeft = {
				colorPoints[H_Kinect.JointType_SpineShoulder],
				colorPoints[H_Kinect.JointType_ShoulderLeft],
				colorPoints[H_Kinect.JointType_ElbowLeft],
				colorPoints[H_Kinect.JointType_WristLeft],
				colorPoints[H_Kinect.JointType_HandLeft],
				colorPoints[H_Kinect.JointType_HandTipLeft]};
		
		if (setPolyline(handLeft, pointx, pointy)) {
			grphOff.drawPolyline(pointx, pointy, handLeft.length);
		}
		
		final Struct_ColorSpacePoint[] thumbLeft = {
				colorPoints[H_Kinect.JointType_HandLeft],
				colorPoints[H_Kinect.JointType_ThumbLeft]};
		
		if (setPolyline(thumbLeft, pointx, pointy)) {
			grphOff.drawPolyline(pointx, pointy, thumbLeft.length);
		}
		
		final Struct_ColorSpacePoint[] footRight = {
				colorPoints[H_Kinect.JointType_SpineBase],
				colorPoints[H_Kinect.JointType_HipRight],
				colorPoints[H_Kinect.JointType_KneeRight],
				colorPoints[H_Kinect.JointType_AnkleRight],
				colorPoints[H_Kinect.JointType_FootRight]};
		
		if (setPolyline(footRight, pointx, pointy)) {
			grphOff.drawPolyline(pointx, pointy, footRight.length);
		}
		
		final Struct_ColorSpacePoint[] footLeft = {
				colorPoints[H_Kinect.JointType_SpineBase],
				colorPoints[H_Kinect.JointType_HipLeft],
				colorPoints[H_Kinect.JointType_KneeLeft],
				colorPoints[H_Kinect.JointType_AnkleLeft],
				colorPoints[H_Kinect.JointType_FootLeft]};
		
		if (setPolyline(footLeft, pointx, pointy)) {
			grphOff.drawPolyline(pointx, pointy, footLeft.length);
		}
		
	}

	private void terminate() throws Exception_KINECT {
		for (int i = 0; i < wib.length; i++) {
			if (wiffrslt[i] != null) wiffrslt[i].Release();
			if (wiff[i] != null) wiff[i].Release();
			if (wiffr[i] != null) wiffr[i].Release();
			if (wiffs[i] != null) wiffs[i].Release();
			
			if (wib[i] != null) wib[i].Release();
		}
		if (wicf != null) wicf.Release();
		if (wicfr != null) wicfr.Release();
		if (wifd != null) wifd.Release();
		if (wicm != null) wicm.Release();
		if (wicfs != null) wicfs.Release();
		if (wibf != null) wibf.Release();
		if (wibfr != null) wibfr.Release();
		if (wibfs != null) wibfs.Release();
		
		if (wiks != null) {
			wiks.Close();
			wiks.Release();
		}
		
		if (grphOff != null) grphOff.dispose();
		threadPool.shutdown();
	}

}
