package testkinectv2;

import javax.swing.*;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.Collection;
import java.util.LinkedList;
import java.util.concurrent.*;
import java.io.*;

import utildll.*;

public class TestKINECTV2_Fusion extends JFrame implements Runnable, ActionListener, ComponentListener {

	// Kinect color, depth and fusion 
	private long[] sensor = {0};
	private long[] source = {0};
	private long[] reader = {0};
	private long[] descript = {0};
	private long[] mapper = {0};
	private long[] frame = {0};
	private long[] sourceD = {0};
	private long[] readerD = {0};
	private long[] descriptD = {0};
	private long[] frameD = {0};
	private long[] reconst = {0};
	private long[] mesh = {0};

	private int[] color = null;

	private int width = 0, height = 0;
	private int widthD = 0, heightD = 0;
	
	private Struct_NUI_FUSION_RECONSTRUCTION_PARAMETERS reconstPrm 
			= new Struct_NUI_FUSION_RECONSTRUCTION_PARAMETERS(256f, 512, 384, 512);
	
	private Struct_Matrix4 worldToCameraTransform = new Struct_Matrix4(1f, 0f, 0f, 0f,
																0f, 1f, 0f, 0f,
																0f, 0f, 1f, 0f,
																0f, 0f, 0f, 1f);
	
	private Struct_NUI_FUSION_CAMERA_PARAMETERS cameraPrm = new Struct_NUI_FUSION_CAMERA_PARAMETERS(
			H_NuiKinectFusionDepthProcessor.NUI_KINECT_DEPTH_NORM_FOCAL_LENGTH_X,
			H_NuiKinectFusionDepthProcessor.NUI_KINECT_DEPTH_NORM_FOCAL_LENGTH_Y,
			H_NuiKinectFusionDepthProcessor.NUI_KINECT_DEPTH_NORM_PRINCIPAL_POINT_X,
			H_NuiKinectFusionDepthProcessor.NUI_KINECT_DEPTH_NORM_PRINCIPAL_POINT_Y);
	
	private long[] depthFloatFrame = {0}, smoothDepthFloatFrame = {0},
			fusionColorFrame = {0}, pointCloudFrame = {0}, surfaceFrame = {0};
	private int[] fusionColor;
	private int[] surface;
	private int errorCount = 0, resetCount = 0;
	private final String MESH_FILE = "mesh.ply";

	// thread pool
	private ExecutorService threadPool = Executors.newFixedThreadPool(8);
	private Collection<Callable<Void>> processes = new LinkedList<Callable<Void>>();
	private final int DEPTH_INDEX_STEP = 6784;
	
	// wrapper
	private Wrp_KINECTV2 wknct = new Wrp_KINECTV2();
	private Wrp_IKinectSensor wiks = null;
	private Wrp_IColorFrameSource wicfs = null;
	private Wrp_IColorFrameReader wicfr = null;
	private Wrp_IFrameDescription wifd = null;
	private Wrp_ICoordinateMapper wicm = null;
	private Wrp_IColorFrame wicf = null;
	private Wrp_IDepthFrameSource widfs = null;
	private Wrp_IDepthFrameReader widfr = null;
	private Wrp_IFrameDescription wifdd = null;
	private Wrp_IDepthFrame widf = null;

	private Wrp_KinectFusion wkf = new Wrp_KinectFusion();
	private Wrp_INuiFusionColorReconstruction winfcr = null;
	private Wrp_JInvoker wjinv = new Wrp_JInvoker();

	// GUI
	private final int IMG_WIDTH = 512;
	private final int IMG_HEIGHT = 424;
	private final int BORDER_WIDTH = 4;
	private final int TITLE_HEIGHT = 30;
	private final int PANEL_HEIGHT = 40;
	private JButton btnSave, btnReset, btnExit;
	private boolean blSave = false, blReset = false, blExit = false; 
	private JProgressBar PrgBar;
	private JLabel lblPrg, lblCount;
	private JPanel pnl;
	private ImageCanvas imgcvs;
	private BufferedImage bfimgOff = null;
	private Graphics2D grphOff = null;
	private boolean blPaint = false;
	
	
	public static void main(String[] args) {
		TestKINECTV2_Fusion testkinectv2 = new TestKINECTV2_Fusion();
	}

	TestKINECTV2_Fusion() {
		this.setTitle("TestKINECTV2_Fusion");
		this.setSize(IMG_WIDTH + BORDER_WIDTH * 2, IMG_HEIGHT + TITLE_HEIGHT + PANEL_HEIGHT + BORDER_WIDTH);
		this.addComponentListener(this);
		this.setLocationRelativeTo(null);
		this.setLayout(null);
		this.setResizable(true);
		getContentPane().setBackground(Color.DARK_GRAY);
		
		imgcvs = new ImageCanvas();
		imgcvs.setBounds(0, 0, IMG_WIDTH, IMG_HEIGHT);
		getContentPane().add(imgcvs);
		
		pnl = new JPanel();
		pnl.setLayout(new FlowLayout(FlowLayout.RIGHT, 10, 5));
		pnl.setBounds(0, IMG_HEIGHT, IMG_WIDTH, PANEL_HEIGHT);
		
		PrgBar = new JProgressBar();
		pnl.add(PrgBar);
		
		lblPrg = new JLabel();
		pnl.add(lblPrg);
		
		btnSave = new JButton("Save Mesh File");
		btnSave.addActionListener(this);
		pnl.add(btnSave);

		lblCount = new JLabel();
		pnl.add(lblCount);
		
		btnReset = new JButton(" Reset ");
		btnReset.addActionListener(this);
		pnl.add(btnReset);

		btnExit = new JButton("  Exit  ");
		btnExit.addActionListener(this);
		pnl.add(btnExit);
		
		getContentPane().add(pnl);
		
		this.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent ev) {
				System.exit(0);
			}
		});

		this.setVisible(true);
		
		Thread thread = new Thread(this);
		thread.start();
	}

	public void actionPerformed(ActionEvent e) {
		JButton btn = (JButton)e.getSource();
		
		if (btn == btnSave) {
			blSave = true;
		}
		if (btn == btnReset) {
			blReset = true;
		}
		if (btn == btnExit) {
			blExit = true;
		}
	}

	public void componentResized(ComponentEvent e) {
		Rectangle rec = e.getComponent().getBounds();
		rec.width -= BORDER_WIDTH * 2;
		rec.height -= TITLE_HEIGHT + PANEL_HEIGHT + BORDER_WIDTH;
		
		System.out.println("componentResized width: " + Integer.toString(rec.width)
				+ " height: " + Integer.toString(rec.height));

		if ((double)IMG_HEIGHT / (double)IMG_WIDTH >= (double)rec.height / (double)rec.width) {
			int w = (int)((double)rec.height * (double)IMG_WIDTH / (double)IMG_HEIGHT);
			imgcvs.setBounds((rec.width - w) / 2, 0, w, rec.height);
		}
		else {
			int h = (int)((double)rec.width * (double)IMG_HEIGHT / (double)IMG_WIDTH);
			imgcvs.setBounds(0, (rec.height - h) / 2, rec.width, h);
		}
		
		pnl.setBounds(0, rec.height, rec.width, PANEL_HEIGHT);
	}

	public void componentMoved(ComponentEvent e) {
	}

	public void componentShown(ComponentEvent e) {
	}

	public void componentHidden(ComponentEvent e) {
	}

	private class ImageCanvas extends Canvas {
		public void paint(Graphics g){
			if (blPaint == true) {
				g.drawImage(bfimgOff, 0, 0, imgcvs.getWidth(), imgcvs.getHeight(), this);
				
				blPaint = false;
			}
		}
		
		public void update(Graphics g){
			paint(g);
		}
	}

	public void run() {
		try {
			initialize();
			drawcvs();
		}
		catch (Exception_KINECT e) {
			System.out.println("***** Exception_KINECT: " + e.getMessage());
		}
		catch (Exception_JInvoker e) {
			System.out.println("***** Exception_JInvoker: " + e.getMessage());
		}
		finally {
			System.out.println("finally start ");
			try {
				terminate();
			}
			catch (Exception_KINECT e) {
				System.out.println("***** Exception_KINECT: " + e.getMessage());
			}
			System.out.println("finally end ");
		}
		System.exit(0);
	}

	private void initialize() throws Exception_KINECT {
		// sensor
		wknct.GetDefaultKinectSensor(sensor);
		System.out.println("sensor: " + Long.toHexString(sensor[0]));
		
		wiks = new Wrp_IKinectSensor(sensor[0]);
		wiks.Open();
		
		// color source
		wiks.get_ColorFrameSource(source);
		System.out.println("source: " + Long.toHexString(source[0]));
		
		wicfs = new Wrp_IColorFrameSource(source[0]);
		
		// color reader
		wicfs.OpenReader(reader);
		System.out.println("reader: " + Long.toHexString(reader[0]));
		
		wicfr = new Wrp_IColorFrameReader(reader[0]);
		
		// color description
		wicfs.get_FrameDescription(descript);
		System.out.println("descript: " + Long.toHexString(descript[0]));
		
		wifd = new Wrp_IFrameDescription(descript[0]);
		
		int[] pWidth = {0};
		int[] pHeight = {0};
		wifd.get_Width(pWidth);
		wifd.get_Height(pHeight);
		System.out.println("width: " + Integer.toString(pWidth[0])
				+ " height: " + Integer.toString(pHeight[0]));
		width = pWidth[0];
		height = pHeight[0];
		
		color = new int[width * height];
		
		// depth source
		wiks.get_DepthFrameSource(sourceD);
		System.out.println("sourceD: " + Long.toHexString(sourceD[0]));
		
		widfs = new Wrp_IDepthFrameSource(sourceD[0]);
		
		// depth reader
		widfs.OpenReader(readerD);
		System.out.println("readerD: " + Long.toHexString(readerD[0]));
		
		widfr = new Wrp_IDepthFrameReader(readerD[0]);

		// depth description
		widfs.get_FrameDescription(descriptD);
		System.out.println("descriptD: " + Long.toHexString(descriptD[0]));
		
		wifdd = new Wrp_IFrameDescription(descriptD[0]);
		
		int[] pWidthD = {0};
		int[] pHeightD = {0};
		wifdd.get_Width(pWidthD);
		wifdd.get_Height(pHeightD);
		System.out.println("widthD: " + Integer.toString(pWidthD[0])
				+ " heightD: " + Integer.toString(pHeightD[0]));
		widthD = pWidthD[0];
		heightD = pHeightD[0];
		
		// coordinate mapper
		wiks.get_CoordinateMapper(mapper);
		System.out.println("mapper: " + Long.toHexString(mapper[0]));
		
		wicm = new Wrp_ICoordinateMapper(mapper[0]);
		
		// create fusion instance
		wkf.NuiFusionCreateColorReconstruction(reconstPrm,
				H_NuiKinectFusionColorVolume.NUI_FUSION_RECONSTRUCTION_PROCESSOR_TYPE_AMP,
				-1, worldToCameraTransform, reconst);
		System.out.println("colorReconstruction: " + Long.toHexString(reconst[0]));
		
		winfcr = new Wrp_INuiFusionColorReconstruction(reconst[0]);

		// create depth image frame for fusion
		wkf.NuiFusionCreateImageFrame(H_NuiKinectFusionColorVolume.NUI_FUSION_IMAGE_TYPE_FLOAT,
				widthD, heightD, cameraPrm, depthFloatFrame);
		System.out.println("depthFloatFrame: " + Long.toString(depthFloatFrame[0]));
		
		// create smooth depth image frame for fusion
		wkf.NuiFusionCreateImageFrame(H_NuiKinectFusionColorVolume.NUI_FUSION_IMAGE_TYPE_FLOAT,
				widthD, heightD, cameraPrm, smoothDepthFloatFrame);
		System.out.println("smoothDepthFloatFrame: " + Long.toString(smoothDepthFloatFrame[0]));
		
		// create color image frame for fusion
		wkf.NuiFusionCreateImageFrame(H_NuiKinectFusionColorVolume.NUI_FUSION_IMAGE_TYPE_COLOR,
				widthD, heightD, cameraPrm, fusionColorFrame);
		System.out.println("fusionColorFrame: " + Long.toString(fusionColorFrame[0]));
		
		// create point cloud image frame for fusion
		wkf.NuiFusionCreateImageFrame(H_NuiKinectFusionColorVolume.NUI_FUSION_IMAGE_TYPE_POINT_CLOUD,
				widthD, heightD, cameraPrm, pointCloudFrame);
		System.out.println("pointCloudFrame: " + Long.toString(pointCloudFrame[0]));
		
		// create surface image frame for fusion
		wkf.NuiFusionCreateImageFrame(H_NuiKinectFusionColorVolume.NUI_FUSION_IMAGE_TYPE_COLOR,
				widthD, heightD, cameraPrm, surfaceFrame);
		System.out.println("surfaceFrame: " + Long.toString(surfaceFrame[0]));
		
		fusionColor = new int[widthD * heightD];
		surface = new int[widthD * heightD];
		
	}

	private void drawcvs() throws Exception_KINECT, Exception_JInvoker {
		bfimgOff = new BufferedImage(widthD, heightD, BufferedImage.TYPE_INT_RGB);
		grphOff = (Graphics2D)bfimgOff.getGraphics();
		
		for (;;) {
			
			if (setSurfaceImage()) {
				blPaint = true;
				imgcvs.repaint();
			}
			
			if (blSave) {
				saveMeshFile();
				blSave = false;
			}
			if (blReset) {
				reset();
				blReset = false;
			}
			if (blExit) {
				break;
			}
		}
	}

	private boolean setSurfaceImage() throws Exception_KINECT, Exception_JInvoker {
		boolean blSet = false;
		
		// color frame
		int hr = wicfr.AcquireLatestFrame(frame);
		System.out.println("frame: " + Long.toHexString(frame[0]));
		
		if (H_WinError.SUCCEEDED(hr) && frame[0] != 0) {
			wicf = new Wrp_IColorFrame(frame[0]);
			
			int capacity = color.length * 4;
			wicf.CopyConvertedFrameDataToArray(capacity, color, H_Kinect.ColorImageFormat_Bgra);
			
			// get depth data
			short[] depthData = getDepthData();
			
			if (depthData != null) {
				
				// translate depth to depth float for fusion
				winfcr.DepthToDepthFloatFrame(depthData, depthData.length * 2, depthFloatFrame[0],
						H_NuiKinectFusionDepthProcessor.NUI_FUSION_DEFAULT_MINIMUM_DEPTH,
						H_NuiKinectFusionDepthProcessor.NUI_FUSION_DEFAULT_MAXIMUM_DEPTH, true);
				
				// smooth depth
				winfcr.SmoothDepthFloatFrame(depthFloatFrame[0], smoothDepthFloatFrame[0],
						H_NuiKinectFusionDepthProcessor.NUI_FUSION_DEFAULT_SMOOTHING_KERNEL_WIDTH,
						H_NuiKinectFusionDepthProcessor.NUI_FUSION_DEFAULT_SMOOTHING_DISTANCE_THRESHOLD);
				
				// get color points for depth frame
				float[] colorPoints = new float[depthData.length * 2]; // ColorSpacePoint {float X, float Y;}
				wicm.MapDepthFrameToColorSpace(depthData.length, depthData, depthData.length, colorPoints);
				
				Struct_NUI_FUSION_IMAGE_FRAME snfif = new Struct_NUI_FUSION_IMAGE_FRAME();
				wjinv.movePtrToStruct(snfif, fusionColorFrame[0]);
				Struct_NUI_FUSION_BUFFER snfb = snfif.pFrameBuffer;
				System.out.println("fusionColorFrame pBits: " + Long.toHexString(snfb.pBits));
				
				for (int ii = 0; ii < depthData.length; ii += DEPTH_INDEX_STEP) {
					final int start = ii;
					final int end = start + DEPTH_INDEX_STEP;
					final float[] colorPointsFinal = colorPoints;
					
					processes.add(new Callable<Void>() {
				        @Override
				        public Void call() throws Exception_KINECT {
					
							for (int i = start; i < end; i++) {
								if (chkFloat(colorPointsFinal[i * 2]) && chkFloat(colorPointsFinal[i * 2 + 1])) {
									int pointX = (int)colorPointsFinal[i * 2];
									int pointY = (int)colorPointsFinal[i * 2 + 1];
									
									if (chkRange(pointX, pointY)) {
										int index = pointX + pointY * width;
										fusionColor[i] = color[index];
									}
									else {
										fusionColor[i] = 0xffff0000; // red in case of range over
									}
								}
								else {
									fusionColor[i] = 0xff00ff00; // green in case of chkFloat error
								}
							}
				            return null;
				        }
				    });
							
				}
				
				try {
				    threadPool.invokeAll(processes);
				} catch (InterruptedException e) {
				    throw new RuntimeException(e);
				} finally {
					processes.clear();
				}

				// set color image frame for fusion
				wjinv.moveIntArrayToPtr(snfb.pBits, fusionColor);
				
				// process depth and color frame
				winfcr.GetCurrentWorldToCameraTransform(worldToCameraTransform);
				hr = winfcr.ProcessFrame(smoothDepthFloatFrame[0], fusionColorFrame[0],
						(short)H_NuiKinectFusionDepthProcessor.NUI_FUSION_DEFAULT_ALIGN_ITERATION_COUNT,
						(short)H_NuiKinectFusionDepthProcessor.NUI_FUSION_DEFAULT_INTEGRATION_WEIGHT,
						H_NuiKinectFusionDepthProcessor.NUI_FUSION_DEFAULT_COLOR_INTEGRATION_OF_ALL_ANGLES,
						null, worldToCameraTransform);
				
				if (H_WinError.FAILED(hr)) {
					errorCount++;
					if (errorCount >= 100) {
						errorCount = 0;
						reset();
					}
				}
				
				// display count
				lblCount.setText("ResetCount:" + Integer.toString(resetCount)
						+ " " + "ErrorCount:" + Integer.toString(errorCount));
				
				// calculate point cloud
				winfcr.CalculatePointCloud(pointCloudFrame[0], surfaceFrame[0], worldToCameraTransform);
				
				// get surface image
				wjinv.movePtrToStruct(snfif, surfaceFrame[0]);
				snfb = snfif.pFrameBuffer;
				System.out.println("surfaceFrame pBits: " + Long.toHexString(snfb.pBits));
				wjinv.movePtrToIntArray(surface, snfb.pBits);

				// set surface image
				bfimgOff.setRGB(0, 0, widthD, heightD, surface, 0, widthD);

				blSet = true;
			}
		}
		
		if (wicf != null) wicf.Release();
		wicf = null;
		frame[0] = 0;
		return blSet;
	}

	private short[] getDepthData() throws Exception_KINECT {
		short[] depthData = null;
		
		// depth frame
		int hr = widfr.AcquireLatestFrame(frameD);
		System.out.println("frameD: " + Long.toHexString(frameD[0]));
		
		if (H_WinError.SUCCEEDED(hr) && frameD[0] != 0) {
			widf = new Wrp_IDepthFrame(frameD[0]);
			
			int[] capacity = {0};
			depthData = widf.AccessUnderlyingBuffer(capacity);
			System.out.println("AccessUnderlyingBuffer capacity: " + Integer.toString(capacity[0])
					+ " buffer.length: " + Integer.toString(depthData.length));
			
		}
		
		if (widf != null) widf.Release();
		widf = null;
		frameD[0] = 0;
		return depthData;
	}
	
	private void saveMeshFile() {
		Wrp_INuiFusionColorMesh winfcm = null;
		long[] pVertices = {0}, pTriangleIndices = {0}, pColors = {0};
		float[] vertices;
		int[] triangleIndices, colors;
		PrintWriter pw = null;
		int curprg = 0, prg = 0;
		
		try {
			// calculate mesh
			winfcr.CalculateMesh(1, mesh);
			System.out.println("color mesh adrs: " + Long.toHexString(mesh[0]));
			winfcm = new Wrp_INuiFusionColorMesh(mesh[0]);
			
			// get counts
			int numVertices = winfcm.VertexCount();
			int numTriangleIndices = winfcm.TriangleVertexIndexCount();
			int numTriangles = numVertices / 3;
			int numColors = winfcm.ColorCount();
			System.out.println("numVertices: " + Integer.toString(numVertices)
					+ " numTriangleIndices: " + Integer.toString(numTriangleIndices)
					+ " numTriangles: " + Integer.toString(numTriangles)
					+ " numColors: " + Integer.toString(numColors));
			
			if (numVertices == 0 || numTriangleIndices == 0 || numVertices % 3 != 0
					|| numTriangleIndices != numVertices || numColors != numVertices) {
				System.out.println("***** vertices number error");
			}
			else {	
				// get vertices
				winfcm.GetVertices(pVertices);
				vertices = new float[numVertices * 3]; // Vector3 {FLOAT x; FLOAT y; FLOAT z;}
				wjinv.movePtrToFloatArray(vertices, pVertices[0]);
				
				// get triangle indices
				winfcm.GetTriangleIndices(pTriangleIndices);
				triangleIndices = new int[numTriangleIndices];
				wjinv.movePtrToIntArray(triangleIndices, pTriangleIndices[0]);
				
				// get colors
				winfcm.GetColors(pColors);
				colors = new int[numColors];
				wjinv.movePtrToIntArray(colors, pColors[0]);
				
				try {
					pw = new PrintWriter(new BufferedWriter(new FileWriter(MESH_FILE)));
					// header
					pw.println("ply");
					pw.println("format ascii 1.0");
					pw.println("comment file created by Kinect Fusion");
						
					// element vertex num, x, y, z, r, g, b
					pw.println(String.format("element vertex %d", numVertices));
					pw.println("property float x");
					pw.println("property float y");
					pw.println("property float z");
					pw.println("property uchar red");
					pw.println("property uchar green");
					pw.println("property uchar blue");
					
					// element face num
					pw.println(String.format("element face %d", numTriangles));
					pw.println("property list uchar int vertex_index");
					pw.println("end_header");
					
					// triangle vertices
					for (int i = 0, idx = 0; i < numVertices; i++, idx += 3) {
						pw.println(String.format("%f %f %f %d %d %d",
								vertices[idx], -vertices[idx + 1], -vertices[idx + 2],
								(colors[i] >> 16) & 0xff, (colors[i] >> 8) & 0xff, colors[i] & 0xff));
						prg = ((i + 1) * 100) / (numVertices + numTriangles);
						if (prg > curprg) {
							PrgBar.setValue(prg);
							lblPrg.setText(Integer.toString(prg) + "%");
							curprg = prg;
						}
					}
	 
					// triangle vertex indices
					for (int i = 0, idx = 0; i < numTriangles; i++, idx += 3) {
						pw.println(String.format("3 %d %d %d", idx, idx + 1, idx + 2));
						prg = ((i + 1 + numVertices) * 100) / (numVertices + numTriangles);
						if (prg > curprg) {
							PrgBar.setValue(prg);
							lblPrg.setText(Integer.toString(prg) + "%");
							curprg = prg;
						}
					}
				} catch (IOException e) {
					System.out.println("***** IOException: " + e.getMessage());
				 }
				finally {
					pw.close();
				}
			}
		}
		catch (Exception_KINECT e) {
			System.out.println("***** Exception_KINECT: " + e.getMessage());
		}
		catch (Exception_JInvoker e) {
			System.out.println("***** Exception_JInvoker: " + e.getMessage());
		}
		finally {
			try {
				if (winfcm != null) winfcm.Release();
			}
			catch (Exception_KINECT e) {
				System.out.println("***** Exception_KINECT: " + e.getMessage());
			}
		}
	}

	private void reset() throws Exception_KINECT {
		resetCount++;
		errorCount = 0;
		worldToCameraTransform = new Struct_Matrix4(1f, 0f, 0f, 0f,
													0f, 1f, 0f, 0f,
													0f, 0f, 1f, 0f,
													0f, 0f, 0f, 1f);
		winfcr.ResetReconstruction(worldToCameraTransform, null);
		System.out.println("resetCount: " + Integer.toString(resetCount));
	}

	private boolean chkFloat(float f) {
		if (Float.isNaN(f)) return false;
		if (Float.isInfinite(f)) return false;
		return true;
	}
	
	private boolean chkRange(int pointX, int pointY) {
		if (pointX < 0 || width <= pointX) return false;
		if (pointY < 0 || height <= pointY) return false;
		return true;
	}
	
	private void terminate() throws Exception_KINECT {
		if (wkf != null) {
			if (depthFloatFrame[0] != 0) wkf.NuiFusionReleaseImageFrame(depthFloatFrame[0]);
			if (smoothDepthFloatFrame[0] != 0) wkf.NuiFusionReleaseImageFrame(smoothDepthFloatFrame[0]);
			if (fusionColorFrame[0] != 0) wkf.NuiFusionReleaseImageFrame(fusionColorFrame[0]);
			if (pointCloudFrame[0] != 0) wkf.NuiFusionReleaseImageFrame(pointCloudFrame[0]);
			if (surfaceFrame[0] != 0) wkf.NuiFusionReleaseImageFrame(surfaceFrame[0]);
		}
		if (winfcr != null) winfcr.Release();
		
		if (wicf != null) wicf.Release();
		if (wicfr != null) wicfr.Release();
		if (wifd != null) wifd.Release();
		if (wicfs != null) wicfs.Release();
		if (wicm != null) wicm.Release();
		if (widf != null) widf.Release();
		if (widfr != null) widfr.Release();
		if (wifdd != null) wifdd.Release();
		if (widfs != null) widfs.Release();
		
		if (wiks != null) {
			wiks.Close();
			wiks.Release();
		}
		
		if (grphOff != null) grphOff.dispose();
		threadPool.shutdown();
	}

}
