package testkinectv2;

import javax.swing.*;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.Collection;
import java.util.LinkedList;
import java.util.concurrent.*;

public class TestKINECTV2_Gesture extends JFrame implements Runnable, ActionListener, ComponentListener {

	// Kinect color, body and gesture
	private long[] sensor = {0};
	private long[] source = {0};
	private long[] reader = {0};
	private long[] descript = {0};
	private long[] mapper = {0};
	private long[] frame = {0};
	private long[] sourceB = {0};
	private long[] readerB = {0};
	private long[] frameB = {0};
	private long[] bodies = new long[H_Kinect.BODY_COUNT];
	private long[][] sourceG = new long[H_Kinect.BODY_COUNT][1];
	private long[][] readerG = new long[H_Kinect.BODY_COUNT][1];
	private long[] database = {0};
	private long[] gestures = null;
	private long[][] frameG = new long[H_Kinect.BODY_COUNT][1];
	private long[][] rsltDiscrete = new long[H_Kinect.BODY_COUNT][1];
	private long[][] rsltContinuous = new long[H_Kinect.BODY_COUNT][1];

	private int[] color = null;
	private int width = 0, height = 0;
	private String[] strGesture = new String[H_Kinect.BODY_COUNT];
	private long[] trackingIdSave = new long[H_Kinect.BODY_COUNT];
	private final String GESTURE_DB_PATH = "YOUNGMAN.gbd";
	//private final String GESTURE_DB_PATH = "HandUp.gba";
	//private final String GESTURE_DB_PATH = "Swipe.gba";
	//private final String GESTURE_DB_PATH = "SampleDatabase.gbd";
	
	private String[] gestureName = null;
	private int[] gestureType = null;

	// thread pool
	private ExecutorService threadPool = Executors.newFixedThreadPool(H_Kinect.BODY_COUNT);
	private Collection<Callable<Void>> processes = new LinkedList<Callable<Void>>();

	// wrapper
	private Wrp_KINECTV2 wknct = new Wrp_KINECTV2();
	private Wrp_IKinectSensor wiks = null;
	private Wrp_IColorFrameSource wicfs = null;
	private Wrp_IColorFrameReader wicfr = null;
	private Wrp_IFrameDescription wifd = null;
	private Wrp_ICoordinateMapper wicm = null;
	private Wrp_IColorFrame wicf = null;
	private Wrp_IBodyFrameSource wibfs = null;
	private Wrp_IBodyFrameReader wibfr = null;
	private Wrp_IBodyFrame wibf = null;
	private Wrp_IBody[] wib = new Wrp_IBody[H_Kinect.BODY_COUNT];
	private Wrp_KINECTV2GESTURE wknctg = new Wrp_KINECTV2GESTURE();
	private Wrp_IVisualGestureBuilderFrameSource[] wivgbfs = new Wrp_IVisualGestureBuilderFrameSource[H_Kinect.BODY_COUNT];
	private Wrp_IVisualGestureBuilderFrameReader[] wivgbfr = new Wrp_IVisualGestureBuilderFrameReader[H_Kinect.BODY_COUNT];
	private Wrp_IVisualGestureBuilderDatabase wivgbd = null;
	private Wrp_IGesture[] wig = null;
	private Wrp_IVisualGestureBuilderFrame[] wivgbf = new Wrp_IVisualGestureBuilderFrame[H_Kinect.BODY_COUNT];
	private Wrp_IDiscreteGestureResult[] widgr = new Wrp_IDiscreteGestureResult[H_Kinect.BODY_COUNT];
	private Wrp_IContinuousGestureResult[] wicgr = new Wrp_IContinuousGestureResult[H_Kinect.BODY_COUNT];
	
	// GUI
	private final int IMG_WIDTH = 1920 / 2;
	private final int IMG_HEIGHT = 1080 / 2;
	private final int BORDER_WIDTH = 4;
	private final int TITLE_HEIGHT = 30;
	private final int PANEL_HEIGHT = 40;
	private final int PANEL_BODY_HEIGHT = 120;
	private JCheckBox chkBody;
	private JButton btnExit;
	private boolean blExit = false; 
	private JPanel pnl;
	private JLabel[] lblGesture = new JLabel[H_Kinect.BODY_COUNT];
	private JPanel pnlGesture;
	private ImageCanvas imgcvs;
	private BufferedImage bfimgOff = null;
	private Graphics2D grphOff = null;
	private boolean blPaint = false;
	private final Color[] colorJoint = {Color.RED, Color.GREEN, Color.BLUE, Color.YELLOW, Color.MAGENTA, Color.CYAN};
	private final Color[] colorText = {new Color(128, 64, 64), new Color(64, 128, 64), new Color(64, 64, 128), 
			new Color(128, 128, 64), new Color(128, 64, 128), new Color(64, 128, 128)};
	private final float LINE_WIDTH = 5.0f;
	
	public static void main(String[] args) {
		TestKINECTV2_Gesture testkinectv2 = new TestKINECTV2_Gesture();
	}
	
	TestKINECTV2_Gesture() {
		this.setTitle("TestKINECTV2_Gesture");
		this.setSize(IMG_WIDTH + BORDER_WIDTH * 2, IMG_HEIGHT + TITLE_HEIGHT + PANEL_BODY_HEIGHT + PANEL_HEIGHT + BORDER_WIDTH);
		this.addComponentListener(this);
		this.setLocationRelativeTo(null);
		this.setLayout(null);
		this.setResizable(true);
		getContentPane().setBackground(Color.DARK_GRAY);
		
		imgcvs = new ImageCanvas();
		imgcvs.setBounds(0, 0, IMG_WIDTH, IMG_HEIGHT);
		getContentPane().add(imgcvs);
		
		pnlGesture = new JPanel();
		pnlGesture.setLayout(new GridLayout(H_Kinect.BODY_COUNT, 1));
		pnlGesture.setBounds(0, IMG_HEIGHT, IMG_WIDTH, PANEL_BODY_HEIGHT);
		
		for (int i = 0; i < lblGesture.length; i++) {
			lblGesture[i] = new JLabel();
			lblGesture[i].setForeground(colorText[i]);
			lblGesture[i].setFont(new Font(Font.MONOSPACED, Font.PLAIN , 12));
			pnlGesture.add(lblGesture[i]);
		}
		
		getContentPane().add(pnlGesture);

		pnl = new JPanel();
		pnl.setLayout(new FlowLayout(FlowLayout.RIGHT, 10, 5));
		pnl.setBounds(0, IMG_HEIGHT + PANEL_BODY_HEIGHT, IMG_WIDTH, PANEL_HEIGHT);
		
		chkBody = new JCheckBox("Body", true);
		pnl.add(chkBody);
		
		btnExit = new JButton("  Exit  ");
		btnExit.addActionListener(this);
		pnl.add(btnExit);
		
		getContentPane().add(pnl);
		
		this.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent ev) {
				System.exit(0);
			}
		});

		this.setVisible(true);
		
		Thread thread = new Thread(this);
		thread.start();
	}
	
	public void actionPerformed(ActionEvent e) {
		JButton btn = (JButton)e.getSource();
		
		if (btn == btnExit) {
			blExit = true;
		}
	}

	public void componentResized(ComponentEvent e) {
		Rectangle rec = e.getComponent().getBounds();
		rec.width -= BORDER_WIDTH * 2;
		rec.height -= TITLE_HEIGHT + PANEL_BODY_HEIGHT + PANEL_HEIGHT + BORDER_WIDTH;
		
		System.out.println("componentResized width: " + Integer.toString(rec.width)
				+ " height: " + Integer.toString(rec.height));

		if ((double)IMG_HEIGHT / (double)IMG_WIDTH >= (double)rec.height / (double)rec.width) {
			int w = (int)((double)rec.height * (double)IMG_WIDTH / (double)IMG_HEIGHT);
			imgcvs.setBounds((rec.width - w) / 2, 0, w, rec.height);
		}
		else {
			int h = (int)((double)rec.width * (double)IMG_HEIGHT / (double)IMG_WIDTH);
			imgcvs.setBounds(0, (rec.height - h) / 2, rec.width, h);
		}
		
		pnlGesture.setBounds(0, rec.height, rec.width, PANEL_BODY_HEIGHT);
		pnl.setBounds(0, rec.height + PANEL_BODY_HEIGHT, rec.width, PANEL_HEIGHT);
	}

	public void componentMoved(ComponentEvent e) {
	}

	public void componentShown(ComponentEvent e) {
	}

	public void componentHidden(ComponentEvent e) {
	}

	private class ImageCanvas extends Canvas {
		public void paint(Graphics g){
			if (blPaint == true) {
				g.drawImage(bfimgOff, 0, 0, imgcvs.getWidth(), imgcvs.getHeight(), this);
				
				blPaint = false;
			}
		}
		
		public void update(Graphics g){
			paint(g);
		}
	}

	public void run() {
		try {
			initialize();
			drawcvs();
		}
		catch (Exception_KINECT e) {
			System.out.println("***** Exception_KINECT: " + e.getMessage());
		}
		finally {
			System.out.println("finally start ");
			try {
				terminate();
			}
			catch (Exception_KINECT e) {
				System.out.println("***** Exception_KINECT: " + e.getMessage());
			}
			System.out.println("finally end ");
		}
		System.exit(0);
	}
	
	private void initialize() throws Exception_KINECT {
		// sensor
		wknct.GetDefaultKinectSensor(sensor);
		System.out.println("sensor: " + Long.toHexString(sensor[0]));
		
		wiks = new Wrp_IKinectSensor(sensor[0]);
		wiks.Open();
		
		// color source
		wiks.get_ColorFrameSource(source);
		System.out.println("source: " + Long.toHexString(source[0]));
		
		wicfs = new Wrp_IColorFrameSource(source[0]);
		
		// color reader
		wicfs.OpenReader(reader);
		System.out.println("reader: " + Long.toHexString(reader[0]));
		
		wicfr = new Wrp_IColorFrameReader(reader[0]);
		
		// color description
		wicfs.get_FrameDescription(descript);
		System.out.println("descript: " + Long.toHexString(descript[0]));
		
		wifd = new Wrp_IFrameDescription(descript[0]);
		
		int[] pWidth = {0};
		int[] pHeight = {0};
		wifd.get_Width(pWidth);
		wifd.get_Height(pHeight);
		System.out.println("width: " + Integer.toString(pWidth[0])
				+ " height: " + Integer.toString(pHeight[0]));
		width = pWidth[0];
		height = pHeight[0];
		
		color = new int[width * height];
		
		// body source
		wiks.get_BodyFrameSource(sourceB);
		System.out.println("sourceB: " + Long.toHexString(sourceB[0]));
		
		wibfs = new Wrp_IBodyFrameSource(sourceB[0]);
		
		// body reader
		wibfs.OpenReader(readerB);
		System.out.println("readerB: " + Long.toHexString(readerB[0]));
		
		wibfr = new Wrp_IBodyFrameReader(readerB[0]);
		
		// coordinate mapper
		wiks.get_CoordinateMapper(mapper);
		System.out.println("mapper: " + Long.toHexString(mapper[0]));
		
		wicm = new Wrp_ICoordinateMapper(mapper[0]);
		
		for (int i = 0; i < H_Kinect.BODY_COUNT; i++) {
			// gesture source
			wknctg.CreateVisualGestureBuilderFrameSource(sensor[0], 0, sourceG[i]);
			System.out.println("sourceG#" + Integer.toString(i) + ": " + Long.toHexString(sourceG[i][0]));
		
			wivgbfs[i] = new Wrp_IVisualGestureBuilderFrameSource(sourceG[i][0]);
			
			// gesture reader
			wivgbfs[i].OpenReader(readerG[i]);
			System.out.println("readerG#" + Integer.toString(i) + ": " + Long.toHexString(readerG[i][0]));
			
			wivgbfr[i] = new Wrp_IVisualGestureBuilderFrameReader(readerG[i][0]);
		}
		
		// create gesture db from file
		wknctg.CreateVisualGestureBuilderDatabaseInstanceFromFile(GESTURE_DB_PATH, database);
		System.out.println("database: " + Long.toHexString(database[0]));
		
		wivgbd = new Wrp_IVisualGestureBuilderDatabase(database[0]);
		
		// get gesture count
		int[] gestureCount = {0};
		wivgbd.get_AvailableGesturesCount(gestureCount);
		System.out.println("gestureCount: " + Long.toString(gestureCount[0]));
		
		if (gestureCount[0] == 0) throw new Exception_KINECT("gestureCount==0");
		
		// get and add gestures
		gestures = new long[gestureCount[0]];
		wivgbd.get_AvailableGestures(gestures.length, gestures);
		
		wig = new Wrp_IGesture[gestures.length];
		gestureName = new String[gestures.length];
		gestureType = new int[gestures.length];
		
		for (int i = 0; i < gestures.length; i++) {
			System.out.println("gestures#" + Integer.toString(i) + ": " + Long.toHexString(gestures[i]));
			
			wig[i] = new Wrp_IGesture(gestures[i]);
			
			for (int j = 0; j < wivgbfs.length; j++) {
				wivgbfs[j].AddGesture(gestures[i]);
				wivgbfs[j].SetIsEnabled(gestures[i], (byte)1);
			}
			
			gestureName[i] = wig[i].get_Name(H_Windef.MAX_PATH);
			System.out.println("gestureName#" + Integer.toString(i) + ": " + gestureName[i]);
			
			int[] type = {0};
			wig[i].get_GestureType(type);
			gestureType[i] = type[0];
			System.out.println("gestureType#" + Integer.toString(i) + ": " + gestureType[i]);
		}
		
		// initialize array
		for (int i = 0; i < H_Kinect.BODY_COUNT; i++) {
			bodies[i] = 0;
			wib[i] = null;
			
			frameG[i][0] = 0;
			wivgbf[i] = null;
			
			rsltDiscrete[i][0] = 0;
			widgr[i] = null;
			
			rsltContinuous[i][0] = 0;
			wicgr[i] = null;
		}
		
	}
	
	private void drawcvs() throws Exception_KINECT {
		bfimgOff = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
		grphOff = (Graphics2D)bfimgOff.getGraphics();
		
		for (;;) {
			
			if (setColorImage()) {
				blPaint = true;
				imgcvs.repaint();
			}
			
			if (blExit) {
				break;
			}
		}
	}

	private boolean setColorImage() throws Exception_KINECT {
		boolean blSet = false;
		
		// color frame
		frame[0] = 0;
		int hr = wicfr.AcquireLatestFrame(frame);
		System.out.println("frame: " + Long.toHexString(frame[0]));
		
		if (H_WinError.SUCCEEDED(hr) && frame[0] != 0) {
			wicf = new Wrp_IColorFrame(frame[0]);
			
			int capacity = color.length * 4;
			wicf.CopyConvertedFrameDataToArray(capacity, color, H_Kinect.ColorImageFormat_Bgra);
			
			// set color image
			bfimgOff.setRGB(0, 0, width, height, color, 0, width);
			
			// draw bodies
			drawBody();

			blSet = true;
		}
		
		if (wicf != null) wicf.Release();
		wicf = null;
		frame[0] = 0;
		return blSet;
	}
	
	private boolean drawBody() throws Exception_KINECT {
		boolean blSet = false;

		// body frame
		frameB[0] = 0;
		int  hr = wibfr.AcquireLatestFrame(frameB);
		System.out.println("frameB: " + Long.toHexString(frameB[0]));
		
		if (H_WinError.SUCCEEDED(hr) && frameB[0] != 0) {
			wibf = new Wrp_IBodyFrame(frameB[0]);
			
			for (int i = 0; i < bodies.length; i++) {
				bodies[i] = 0;
			}
			wibf.GetAndRefreshBodyData(bodies.length, bodies);
			System.out.println("bodies: " + Long.toHexString(bodies[0])
					+ " " + Long.toHexString(bodies[1]) + " " + Long.toHexString(bodies[2])
					+ " " + Long.toHexString(bodies[3]) + " " + Long.toHexString(bodies[4])
					+ " " + Long.toHexString(bodies[5]));
			
			for (int ii = 0; ii < bodies.length; ii++) {
				final int i = ii;
				processes.add(new Callable<Void>() {
			        @Override
			        public Void call() throws Exception_KINECT {
						strGesture[i] = null;
						wib[i] = null;
						
						if (bodies[i] != 0) {
							wib[i] = new Wrp_IBody(bodies[i]);
							
							byte[] tracked = {0};
							wib[i].get_IsTracked(tracked);
							System.out.println("tracked: " + Byte.toString(tracked[0]));
							
							if (tracked[0] != 0) {
								
								Struct_Joint[] joints = new Struct_Joint[H_Kinect.JointType_Count];
								for (int j = 0; j < joints.length; j++) {
									joints[j] = new Struct_Joint();
								}
								wib[i].GetJoints(joints.length, joints);
								
								Struct_ColorSpacePoint[] colorPoints = new Struct_ColorSpacePoint[joints.length];
								for (int j = 0; j < colorPoints.length; j++) {
									if (joints[j].TrackingState != H_Kinect.TrackingState_NotTracked) {
										if (chkFloat(joints[j].Position.X) && chkFloat(joints[j].Position.Y) && chkFloat(joints[j].Position.Z)) {
											colorPoints[j] = new Struct_ColorSpacePoint();
											wicm.MapCameraPointToColorSpace(joints[j].Position, colorPoints[j]);
										}
									}
								}
								
								if (chkBody.isSelected()) {
									// draw joints
									drawJoints(colorPoints, i);
								}
								
								// set tracking ID
								long[] trackingId = {0};
								wib[i].get_TrackingId(trackingId);
								System.out.println("trackingId: " + Long.toHexString(trackingId[0]));
								if (trackingId[0] != trackingIdSave[i]) {
									wivgbfs[i].put_TrackingId(trackingId[0]);
									trackingIdSave[i] = trackingId[0];
								}
								
								// draw Gesture
								drawGesture(colorPoints, i);
							}
						}
						if (strGesture[i] == null) strGesture[i] = "Body#" + Integer.toString(i) + " NotTracked";
						lblGesture[i].setText(strGesture[i]);
			            return null;
			        }
			    });
						
			}
			
			try {
			    threadPool.invokeAll(processes);
			} catch (InterruptedException e) {
			    throw new RuntimeException(e);
			} finally {
				processes.clear();
			}

			for (int i = 0; i < wib.length; i++) {
				if (wib[i] != null) wib[i].Release();
				wib[i] = null;
				bodies[i] = 0;
			}
			
			blSet = true;
		}
		
		if (wibf != null) wibf.Release();
		wibf = null;
		frameB[0] = 0;
		return blSet;
	}

	private boolean drawGesture(Struct_ColorSpacePoint[] colorPoints, int bodyNo) throws Exception_KINECT {
		boolean blSet = false;
	
		frameG[bodyNo][0] = 0;
		int hr = wivgbfr[bodyNo].CalculateAndAcquireLatestFrame(frameG[bodyNo]);
		System.out.println("frameG: " + Long.toHexString(frameG[bodyNo][0]));
		
		if (H_WinError.SUCCEEDED(hr) && frameG[bodyNo][0] != 0) {
			
			wivgbf[bodyNo] = new Wrp_IVisualGestureBuilderFrame(frameG[bodyNo][0]);
			byte[] isTrackingIdValid = {0};
			wivgbf[bodyNo].get_IsTrackingIdValid(isTrackingIdValid);
			System.out.println("isTrackingIdValid: " + Byte.toString(isTrackingIdValid[0]));
			
			if (isTrackingIdValid[0] != 0) {
				strGesture[bodyNo] = "Body#" + Integer.toString(bodyNo) + " Tracked   ";
			
				for (int i = 0; i < gestures.length; i++) {
					
					if (gestureType[i] == H_Kinect_VisualGestureBuilder.GestureType_Discrete) {
						// get discrete gesture result
						rsltDiscrete[bodyNo][0] = 0;
						wivgbf[bodyNo].get_DiscreteGestureResult(gestures[i], rsltDiscrete[bodyNo]);
						System.out.println("rsltDiscrete#" + Integer.toString(i) + ": " + Long.toHexString(rsltDiscrete[bodyNo][0]));
						
						widgr[bodyNo] = new Wrp_IDiscreteGestureResult(rsltDiscrete[bodyNo][0]);
						
						// detect gesture
						byte[] detected = {0};
						widgr[bodyNo].get_Detected(detected);
						System.out.println("detected#" + Integer.toString(i) + ": " + Byte.toString(detected[0]));
						
						// get confidence
						float[] confidence = {0};
						widgr[bodyNo].get_Confidence(confidence);
						System.out.println("confidence#" + Integer.toString(i) + ": " + Float.toString(confidence[0]));
						
						strGesture[bodyNo] += " Gesture#" + Integer.toString(i) 
								+ " Name:" + gestureName[i]
								+ " Type:Discrete Detected:" + (detected[0] == 0 ? "No " : "Yes")
								+ String.format(" Confidence:%10.5f", confidence[0]); 
						
						if (detected[0] != 0) {
							drawGestureText(gestureName[i], colorPoints[H_Kinect.JointType_Head], bodyNo);
							blSet = true;
						}
						
						if (widgr[bodyNo] != null) widgr[bodyNo].Release();
						widgr[bodyNo] = null;
						rsltDiscrete[bodyNo][0] = 0;
					}
					else if (gestureType[i] == H_Kinect_VisualGestureBuilder.GestureType_Continuous) {
						// get continuous gesture result
						rsltContinuous[bodyNo][0] = 0;
						wivgbf[bodyNo].get_ContinuousGestureResult(gestures[i], rsltContinuous[bodyNo]);
						System.out.println("rsltContinuous#" + Integer.toString(i) + ": " + Long.toHexString(rsltContinuous[bodyNo][0]));
						
						wicgr[bodyNo] = new Wrp_IContinuousGestureResult(rsltContinuous[bodyNo][0]);
						
						// get progress
						float[] progress = {0.0f};
						wicgr[bodyNo].get_Progress(progress);
						System.out.println("progress#" + Integer.toString(i) + ": " + Float.toString(progress[0]));
						
						strGesture[bodyNo] += " Gesture#" + Integer.toString(i)
								+ " Name:" + gestureName[i]
								+ " Type:Continuous"
								+ String.format(" Progress:%10.5f", progress[0]); 
	
						if (progress[0] >= 0.9f) {
							drawGestureText(gestureName[i], colorPoints[H_Kinect.JointType_Head], bodyNo);
							blSet = true;
						}
						
						if (wicgr[bodyNo] != null) wicgr[bodyNo].Release();
						wicgr[bodyNo] = null;
						rsltContinuous[bodyNo][0] = 0;
					}
				}
			}
		}
		
		if (wivgbf[bodyNo] != null) wivgbf[bodyNo].Release();
		wivgbf[bodyNo] = null;
		frameG[bodyNo][0] = 0;
		return blSet;
	}
	
	private void drawGestureText(String str, Struct_ColorSpacePoint colorPoint, int bodyNo) {
		grphOff.setFont(new Font(Font.MONOSPACED, Font.BOLD , 96));
		
		if (colorPoint != null) {
			if (chkFloat(colorPoint.X) && chkFloat(colorPoint.Y)) {
				int pointX = (int)colorPoint.X;
				int pointY = (int)colorPoint.Y;
				
				if (chkRange(pointX, pointY)) {
					grphOff.setColor(Color.WHITE);
					grphOff.drawString(str, colorPoint.X + 2, colorPoint.Y);
					grphOff.drawString(str, colorPoint.X - 2, colorPoint.Y);
					grphOff.drawString(str, colorPoint.X, colorPoint.Y + 2);
					grphOff.drawString(str, colorPoint.X, colorPoint.Y - 2);
					grphOff.setColor(colorJoint[bodyNo]);
					grphOff.drawString(str, colorPoint.X, colorPoint.Y);
				}
			}
		}
	}
	
	private boolean chkFloat(float f) {
		if (Float.isNaN(f)) return false;
		if (Float.isInfinite(f)) return false;
		return true;
	}
	
	private boolean chkRange(int pointX, int pointY) {
		if (pointX < 0 || width <= pointX) return false;
		if (pointY < 0 || height <= pointY) return false;
		return true;
	}
	
	private boolean setPolyline(Struct_ColorSpacePoint[] colorPoints, int[] pointX, int[] pointY) {
		
		for (int i = 0; i < colorPoints.length; i++) {
			if (colorPoints[i] == null) return false;
			if (!chkFloat(colorPoints[i].X) || !chkFloat(colorPoints[i].Y)) return false;
			pointX[i] = (int)colorPoints[i].X;
			pointY[i] = (int)colorPoints[i].Y;
			if (!chkRange(pointX[i], pointY[i])) return false;
		}
	
		return true;
	}
	
	private void drawJoints(Struct_ColorSpacePoint[] colorPoints, int bodyNo) {
		
		grphOff.setStroke(new BasicStroke(LINE_WIDTH));
		grphOff.setColor(colorJoint[bodyNo]);
		final int r = 5;
		final int r2 = r * 2;
		
		// draw joints
		for (int i = 0; i < colorPoints.length; i++) {
			if (colorPoints[i] != null) {
				if (chkFloat(colorPoints[i].X) && chkFloat(colorPoints[i].Y)) {
					int pointX = (int)colorPoints[i].X;
					int pointY = (int)colorPoints[i].Y;
					
					if (chkRange(pointX, pointY)) {
						grphOff.drawOval(pointX - r, pointY - r, r2, r2);
					}
				}
			}
		}
		
		// draw lines
		int[] pointx = new int[6];
		int[] pointy = new int[6];
		
		grphOff.setStroke(new BasicStroke(LINE_WIDTH));
		grphOff.setColor(colorJoint[bodyNo]);

		
		final Struct_ColorSpacePoint[] headSpine = {
				colorPoints[H_Kinect.JointType_Head],
				colorPoints[H_Kinect.JointType_Neck],
				colorPoints[H_Kinect.JointType_SpineShoulder],
				colorPoints[H_Kinect.JointType_SpineMid],
				colorPoints[H_Kinect.JointType_SpineBase]};
		
		if (setPolyline(headSpine, pointx, pointy)) {
			grphOff.drawPolyline(pointx, pointy, headSpine.length);
		}
		
		final Struct_ColorSpacePoint[] handRight = {
				colorPoints[H_Kinect.JointType_SpineShoulder],
				colorPoints[H_Kinect.JointType_ShoulderRight],
				colorPoints[H_Kinect.JointType_ElbowRight],
				colorPoints[H_Kinect.JointType_WristRight],
				colorPoints[H_Kinect.JointType_HandRight],
				colorPoints[H_Kinect.JointType_HandTipRight]};
		
		if (setPolyline(handRight, pointx, pointy)) {
			grphOff.drawPolyline(pointx, pointy, handRight.length);
		}
		
		final Struct_ColorSpacePoint[] thumbRight = {
				colorPoints[H_Kinect.JointType_HandRight],
				colorPoints[H_Kinect.JointType_ThumbRight]};
		
		if (setPolyline(thumbRight, pointx, pointy)) {
			grphOff.drawPolyline(pointx, pointy, thumbRight.length);
		}
		
		final Struct_ColorSpacePoint[] handLeft = {
				colorPoints[H_Kinect.JointType_SpineShoulder],
				colorPoints[H_Kinect.JointType_ShoulderLeft],
				colorPoints[H_Kinect.JointType_ElbowLeft],
				colorPoints[H_Kinect.JointType_WristLeft],
				colorPoints[H_Kinect.JointType_HandLeft],
				colorPoints[H_Kinect.JointType_HandTipLeft]};
		
		if (setPolyline(handLeft, pointx, pointy)) {
			grphOff.drawPolyline(pointx, pointy, handLeft.length);
		}
		
		final Struct_ColorSpacePoint[] thumbLeft = {
				colorPoints[H_Kinect.JointType_HandLeft],
				colorPoints[H_Kinect.JointType_ThumbLeft]};
		
		if (setPolyline(thumbLeft, pointx, pointy)) {
			grphOff.drawPolyline(pointx, pointy, thumbLeft.length);
		}
		
		final Struct_ColorSpacePoint[] footRight = {
				colorPoints[H_Kinect.JointType_SpineBase],
				colorPoints[H_Kinect.JointType_HipRight],
				colorPoints[H_Kinect.JointType_KneeRight],
				colorPoints[H_Kinect.JointType_AnkleRight],
				colorPoints[H_Kinect.JointType_FootRight]};
		
		if (setPolyline(footRight, pointx, pointy)) {
			grphOff.drawPolyline(pointx, pointy, footRight.length);
		}
		
		final Struct_ColorSpacePoint[] footLeft = {
				colorPoints[H_Kinect.JointType_SpineBase],
				colorPoints[H_Kinect.JointType_HipLeft],
				colorPoints[H_Kinect.JointType_KneeLeft],
				colorPoints[H_Kinect.JointType_AnkleLeft],
				colorPoints[H_Kinect.JointType_FootLeft]};
		
		if (setPolyline(footLeft, pointx, pointy)) {
			grphOff.drawPolyline(pointx, pointy, footLeft.length);
		}
		
	}
	
	private void terminate() throws Exception_KINECT {
		if (wivgbd != null) wivgbd.Release();
		
		if (wig != null) {
			for (int i = 0; i < wig.length; i++) {
				if (wig[i] != null) wig[i].Release();
			}
		}
		
		for (int i = 0; i < wib.length; i++) {
			if (widgr[i] != null) widgr[i].Release();
			if (wicgr[i] != null) wicgr[i].Release();
			if (wivgbf[i] != null) wivgbf[i].Release();
			if (wivgbfr[i] != null) wivgbfr[i].Release();
			if (wivgbfs[i] != null) wivgbfs[i].Release();
			
			if (wib[i] != null) wib[i].Release();
		}
		
		if (wicf != null) wicf.Release();
		if (wicfr != null) wicfr.Release();
		if (wifd != null) wifd.Release();
		if (wicm != null) wicm.Release();
		if (wicfs != null) wicfs.Release();
		if (wibf != null) wibf.Release();
		if (wibfr != null) wibfr.Release();
		if (wibfs != null) wibfs.Release();
		
		if (wiks != null) {
			wiks.Close();
			wiks.Release();
		}
		
		if (grphOff != null) grphOff.dispose();
		threadPool.shutdown();
	}
	
}
