package testkinectv2;

import javax.swing.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;

public class TestKINECTV2_Infrared extends JFrame implements Runnable, ActionListener {

	// Kinect infrared and long exposure infrared
	private long[] sensor = {0};
	private long[] source = {0};
	private long[] reader = {0};
	private long[] descript = {0};
	private long[] frame = {0};
	private long[] sourceLE = {0};
	private long[] readerLE = {0};
	private long[] descriptLE = {0};
	private long[] frameLE = {0};
	
	private int[] color = null;
	private int[] colorLE = null;
	private int width = 0, height = 0;
	private int widthLE = 0, heightLE = 0;
	
	// wrapper
	private Wrp_KINECTV2 wknct = new Wrp_KINECTV2();
	private Wrp_IKinectSensor wiks = null;
	private Wrp_IInfraredFrameSource wiifs = null;
	private Wrp_IInfraredFrameReader wiifr = null;
	private Wrp_IFrameDescription wifd = null;
	private Wrp_IInfraredFrame wiif = null;
	private Wrp_ILongExposureInfraredFrameSource wileifs = null;
	private Wrp_ILongExposureInfraredFrameReader wileifr = null;
	private Wrp_IFrameDescription wifdle = null;
	private Wrp_ILongExposureInfraredFrame wileif = null;
	
	// GUI
	private final int IMG_WIDTH = 512;
	private final int IMG_HEIGHT = 424;
	private final int BORDER_WIDTH = 4;
	private final int TITLE_HEIGHT = 30;
	private final int PANEL_HEIGHT = 40;
	private JButton btnExit;
	private boolean blExit = false; 
	private JPanel pnl;
	private ImageCanvas imgcvs;
	private BufferedImage bfimgOff = new BufferedImage(IMG_WIDTH, IMG_HEIGHT, BufferedImage.TYPE_INT_RGB);
	private Graphics grphOff = bfimgOff.getGraphics();
	private boolean blPaint = false;
	private ImageCanvasLE imgcvsLE;
	private BufferedImage bfimgOffLE = new BufferedImage(IMG_WIDTH, IMG_HEIGHT, BufferedImage.TYPE_INT_RGB);
	private Graphics grphOffLE = bfimgOffLE.getGraphics();
	private boolean blPaintLE = false;
	
	public static void main(String[] args) {
		TestKINECTV2_Infrared testkinectv2 = new TestKINECTV2_Infrared();
	}
	
	TestKINECTV2_Infrared() {
		this.setTitle("TestKINECTV2_Infrared");
		this.setSize(IMG_WIDTH * 2 + BORDER_WIDTH * 2, IMG_HEIGHT + TITLE_HEIGHT + PANEL_HEIGHT + BORDER_WIDTH);
		this.setLocationRelativeTo(null);
		this.setLayout(null);
		this.setResizable(false);
		
		imgcvs = new ImageCanvas();
		imgcvs.setBounds(0, 0, IMG_WIDTH, IMG_HEIGHT);
		getContentPane().add(imgcvs);
		
		imgcvsLE = new ImageCanvasLE();
		imgcvsLE.setBounds(IMG_WIDTH, 0, IMG_WIDTH, IMG_HEIGHT);
		getContentPane().add(imgcvsLE);
		
		pnl = new JPanel();
		pnl.setLayout(new FlowLayout(FlowLayout.RIGHT, 10, 5));
		pnl.setBounds(0, IMG_HEIGHT, IMG_WIDTH * 2, PANEL_HEIGHT);
		
		btnExit = new JButton("  Exit  ");
		btnExit.addActionListener(this);
		pnl.add(btnExit);
		
		getContentPane().add(pnl);
		
		this.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent ev) {
				System.exit(0);
			}
		});

		this.setVisible(true);
		
		Thread thread = new Thread(this);
		thread.start();
	}

	public void actionPerformed(ActionEvent e) {
		JButton btn = (JButton)e.getSource();
		
		if (btn == btnExit) {
			blExit = true;
		}
	}

	private class ImageCanvas extends Canvas {
		public void paint(Graphics g){
			if (blPaint == true) {
				g.drawImage(bfimgOff, 0, 0, this);
				blPaint = false;
			}
		}
		
		public void update(Graphics g){
			paint(g);
		}
	}

	private class ImageCanvasLE extends Canvas {
		public void paint(Graphics g){
			if (blPaintLE == true) {
				g.drawImage(bfimgOffLE, 0, 0, this);
				blPaintLE = false;
			}
		}
		
		public void update(Graphics g){
			paint(g);
		}
	}

	public void run() {
		try {
			initialize();
			drawcvs();
		}
		catch (Exception_KINECT e) {
			System.out.println("***** Exception_KINECT: " + e.getMessage());
		}
		finally {
			System.out.println("finally start ");
			try {
				terminate();
			}
			catch (Exception_KINECT e) {
				System.out.println("***** Exception_KINECT: " + e.getMessage());
			}
			System.out.println("finally end ");
		}
		System.exit(0);
	}

	void initialize() throws Exception_KINECT {
		// sensor
		wknct.GetDefaultKinectSensor(sensor);
		System.out.println("sensor: " + Long.toHexString(sensor[0]));
		
		wiks = new Wrp_IKinectSensor(sensor[0]);
		wiks.Open();
		
		// infrared source
		wiks.get_InfraredFrameSource(source);
		System.out.println("source: " + Long.toHexString(source[0]));
		
		wiifs = new Wrp_IInfraredFrameSource(source[0]);
		
		// infrared reader
		wiifs.OpenReader(reader);
		System.out.println("reader: " + Long.toHexString(reader[0]));
		
		wiifr = new Wrp_IInfraredFrameReader(reader[0]);
		
		// infrared description
		wiifs.get_FrameDescription(descript);
		System.out.println("descript: " + Long.toHexString(descript[0]));
		
		wifd = new Wrp_IFrameDescription(descript[0]);
		
		int[] pWidth = {0};
		int[] pHeight = {0};
		wifd.get_Width(pWidth);
		wifd.get_Height(pHeight);
		System.out.println("width: " + Integer.toString(pWidth[0])
				+ " height: " + Integer.toString(pHeight[0]));
		width = pWidth[0];
		height = pHeight[0];
		
		color = new int[width * height];
		
		// long exposure infrared source
		wiks.get_LongExposureInfraredFrameSource(sourceLE);
		System.out.println("sourceLE: " + Long.toHexString(sourceLE[0]));
		
		wileifs = new Wrp_ILongExposureInfraredFrameSource(sourceLE[0]);
		
		// long exposure infrared reader
		wileifs.OpenReader(readerLE);
		System.out.println("readerLE: " + Long.toHexString(readerLE[0]));
		
		wileifr = new Wrp_ILongExposureInfraredFrameReader(readerLE[0]);
		
		// long exposure infrared description
		wileifs.get_FrameDescription(descriptLE);
		System.out.println("descriptLE: " + Long.toHexString(descriptLE[0]));
		
		wifdle = new Wrp_IFrameDescription(descriptLE[0]);
		
		int[] pWidthLE = {0};
		int[] pHeightLE = {0};
		wifdle.get_Width(pWidthLE);
		wifdle.get_Height(pHeightLE);
		System.out.println("widthLE: " + Integer.toString(pWidthLE[0])
				+ " heightLE: " + Integer.toString(pHeightLE[0]));
		widthLE = pWidthLE[0];
		heightLE = pHeightLE[0];
		
		colorLE = new int[widthLE * heightLE];
	}
	
	void drawcvs() throws Exception_KINECT {
		
		for (;;) {
			
			if (setInfraredImage()) {
				blPaint = true;
				imgcvs.repaint();
			}
			
			if (setLongExposureInfraredImage()) {
				blPaintLE = true;
				imgcvsLE.repaint();
			}
			
			if (blExit) {
				break;
			}
		}
	}
	
	boolean setInfraredImage() throws Exception_KINECT {
		boolean blSet = false;
		
		// infrared frame
		frame[0] = 0;
		int hr = wiifr.AcquireLatestFrame(frame);
		System.out.println("frame: " + Long.toHexString(frame[0]));
		
		if (H_WinError.SUCCEEDED(hr)) {
			wiif = new Wrp_IInfraredFrame(frame[0]);
			
			int[] capacity = {0};
			short[] buffer = wiif.AccessUnderlyingBuffer(capacity);
			System.out.println("AccessUnderlyingBuffer capacity: " + Integer.toString(capacity[0])
					+ " buffer.length: " + Integer.toString(buffer.length));
			
			for (int i = 0; i < color.length; i++) {
				int gray = (buffer[i] >> 8) & 0xff;
				color[i] = (gray << 16) | (gray << 8) | gray; 
			}
			
			// set color image
			bfimgOff.setRGB(0, 0, IMG_WIDTH, IMG_HEIGHT, color, 0, IMG_WIDTH);
			
			blSet = true;
		}
		
		if (wiif != null) wiif.Release();
		wiif = null;
		frame[0] = 0;
		return blSet;
	}
	
	boolean setLongExposureInfraredImage() throws Exception_KINECT {
		boolean blSet = false;

		// long exposure infrared frame
		frameLE[0] = 0;
		int  hr = wileifr.AcquireLatestFrame(frameLE);
		System.out.println("frameLE: " + Long.toHexString(frameLE[0]));
		
		if (H_WinError.SUCCEEDED(hr)) {
			wileif = new Wrp_ILongExposureInfraredFrame(frameLE[0]);
			
			int[] capacityLE = {0};
			short[] bufferLE = wileif.AccessUnderlyingBuffer(capacityLE);
			System.out.println("AccessUnderlyingBuffer capacityLE: " + Integer.toString(capacityLE[0])
					+ " bufferLE.length: " + Integer.toString(bufferLE.length));

			for (int i = 0; i < colorLE.length; i++) {
				int gray = (bufferLE[i] >> 8) & 0xff;
				colorLE[i] = (gray << 16) | (gray << 8) | gray; 
			}
			
			// set color image
			bfimgOffLE.setRGB(0, 0, IMG_WIDTH, IMG_HEIGHT, colorLE, 0, IMG_WIDTH);
			
			blSet = true;
		}
		
		if (wileif != null) wileif.Release();
		wileif = null;
		frameLE[0] = 0;
		return blSet;
	}
	
	void terminate() throws Exception_KINECT {
		if (wiif != null) wiif.Release();
		if (wiifr != null) wiifr.Release();
		if (wifd != null) wifd.Release();
		if (wiifs != null) wiifs.Release();
		if (wileif != null) wileif.Release();
		if (wileifr != null) wileifr.Release();
		if (wifdle != null) wifdle.Release();
		if (wileifs != null) wileifs.Release();
		
		if (wiks != null) {
			wiks.Close();
			wiks.Release();
		}
		
		if (grphOff != null) grphOff.dispose();
		if (grphOffLE != null) grphOffLE.dispose();
	}

}
