package testkinectv2;

import utildll.*;

public class Wrp_Kernel32 {
	static final int CHARSET = Arg.ANSI;
	static final String KERNEL32_DLL_NAME = "kernel32";
	private String[] strFncName = {
			"RtlMoveMemory",
			"CreateEventA",
			"ResetEvent",
			"WaitForSingleObject",
			"WaitForMultipleObjects",
			"CloseHandle"};
	private long[] lngFncAdrs = new long[strFncName.length];
	private final int 
		IDX_RtlMoveMemory = 0,
		IDX_CreateEvent = 1,
		IDX_ResetEvent = 2,
		IDX_WaitForSingleObject = 3,
		IDX_WaitForMultipleObjects = 4,
		IDX_CloseHandle = 5;
	private long hMod = 0;
	
	Wrp_Kernel32() {
		Arg_GetModuleHandle agmh = new Arg_GetModuleHandle(KERNEL32_DLL_NAME);
		
		hMod = JInvoker.callLongProc("GetModuleHandle", "kernel32", agmh);
		
		System.out.println("GetModuleHandle rtn: " + Long.toHexString(hMod)
				+ " LastError: " + Integer.toString(agmh.getLastError()));
		
		for (int i = 0; i < strFncName.length; i++) {
			Arg_GetProcAddress agpa = new Arg_GetProcAddress(hMod, strFncName[i]);
			
			lngFncAdrs[i] = JInvoker.callLongProc("GetProcAddress", "kernel32", agpa);
			
			System.out.println("GetProcAddress rtn: " + Long.toHexString(lngFncAdrs[i])
					+ " LastError: " + Integer.toString(agpa.getLastError())
					+ " FncName: " + strFncName[i]);
		}
	}

	class Arg_property {
		private int _LastError;
	
		int getLastError() {return _LastError;}
	}

	class Arg_RtlMoveMemory extends Arg_property {
		private long Destination;
		private long Source;
		private int Length;
		
		Arg_RtlMoveMemory(long Destination, long Source, int Length) {
			this.Destination = Destination;
			this.Source = Source;
			this.Length = Length;
		}
	}

	void RtlMoveMemory(long Destination, long Source, int Length) throws Exception_Win64 {
		Arg_RtlMoveMemory armm = new Arg_RtlMoveMemory(Destination, Source, Length);

		JInvoker.callVoidProcAddress(lngFncAdrs[IDX_RtlMoveMemory], armm);
		
		System.out.println("RtlMoveMemory LastError: " + Integer.toString(armm.getLastError()));
		
		if (armm.getLastError() != 0) throw new Exception_Win64("RtlMoveMemory");
	}

	/*
	HANDLE CreateEvent(
		LPSECURITY_ATTRIBUTES lpEventAttributes,
		BOOL bManualReset,
		BOOL bInitialState,
		LPCTSTR lpName
	);
	 */
	@_CharSet(CHARSET)
	class Arg_CreateEvent extends Arg_property {
		private Object lpEventAttributes;
		private boolean bManualReset;
		private boolean bInitialState;
		private @_Const String lpName;
		
		Arg_CreateEvent(Object lpEventAttributes, boolean bManualReset, boolean bInitialState,
				String lpName) {
			this.lpEventAttributes = lpEventAttributes;
			this.bManualReset = bManualReset;
			this.bInitialState = bInitialState;
			this.lpName = lpName;
		}
	}

	long CreateEvent(Object lpEventAttributes, boolean bManualReset, boolean bInitialState,
			String lpName) throws Exception_Win64 {
		Arg_CreateEvent ace = new Arg_CreateEvent(lpEventAttributes, bManualReset, bInitialState,
				lpName);

		long lngRtn = JInvoker.callLongProcAddress(lngFncAdrs[IDX_CreateEvent], ace);
		
		System.out.println("CreateEvent HANDLE: "  + Long.toHexString(lngRtn)
				+ " LastError: " + Integer.toString(ace.getLastError()));
		
		if (ace.getLastError() != 0) throw new Exception_Win64("CreateEvent");
		if (lngRtn == 0) throw new Exception_Win64("CreateEvent");

		return lngRtn;
	}
	
	/*
	BOOL ResetEvent(
		HANDLE hEvent
	);
	 */
	class Arg_ResetEvent extends Arg_property {
		private long hEvent;
		
		Arg_ResetEvent(long hEvent) {
			this.hEvent = hEvent;
		}
	}

	boolean ResetEvent(long hEvent) throws Exception_Win64 {
		Arg_ResetEvent are = new Arg_ResetEvent(hEvent);

		boolean blRtn = JInvoker.callBooleanProcAddress(lngFncAdrs[IDX_ResetEvent], are);
		
		System.out.println("ResetEvent ret: "  + Boolean.toString(blRtn)
				+ " LastError: " + Integer.toString(are.getLastError()));
		
		if (are.getLastError() != 0) throw new Exception_Win64("ResetEvent");
		if (blRtn == false) throw new Exception_Win64("ResetEvent");

		return blRtn;
	}
	
	/*
	DWORD WaitForSingleObject(
		HANDLE hHandle,
		DWORD dwMilliseconds
	);
	 */
	class Arg_WaitForSingleObject extends Arg_property {
		private long hHandle;
		private int dwMilliseconds;
		
		Arg_WaitForSingleObject(long hHandle, int dwMilliseconds) {
			this.hHandle = hHandle;
			this.dwMilliseconds = dwMilliseconds;
		}
	}
	
	int WaitForSingleObject(long hHandle, int dwMilliseconds) throws Exception_Win64 {
		Arg_WaitForSingleObject awfso = new Arg_WaitForSingleObject(hHandle, dwMilliseconds);

		int intRtn = JInvoker.callIntProcAddress(lngFncAdrs[IDX_WaitForSingleObject], awfso);
		
		System.out.println("WaitForSingleObject rtn: "  + Integer.toHexString(intRtn)
				+ " LastError: " + Integer.toString(awfso.getLastError()));
		
		if (awfso.getLastError() != 0) throw new Exception_Win64("WaitForSingleObject");
		if (intRtn == H_WinBase.WAIT_FAILED) throw new Exception_Win64("WaitForSingleObject");

		return intRtn;
	}

	/*
	DWORD WaitForMultipleObjects(
		DWORD nCount,             // z̃nh̐
		CONST HANDLE *lpHandles,  // IuWFNgnhȂz
		BOOL fWaitAll,            // ҋ@IvV
		DWORD dwMilliseconds      // ^CAEg
	);
	 */
	class Arg_WaitForMultipleObjects extends Arg_property {
		private int nCount;
		private long[] lpHandles;
		private boolean fWaitAll;
		private int dwMilliseconds;
		
		Arg_WaitForMultipleObjects(int nCount, long[] lpHandles, boolean fWaitAll, 
				int dwMilliseconds) {
			this.nCount = nCount;
			this.lpHandles = lpHandles;
			this.fWaitAll = fWaitAll;
			this.dwMilliseconds = dwMilliseconds;
		}
	}
	
	int WaitForMultipleObjects(int nCount, long[] lpHandles, boolean fWaitAll, int dwMilliseconds)
			throws Exception_Win64 {
		Arg_WaitForMultipleObjects awfmo = new Arg_WaitForMultipleObjects(nCount, lpHandles,
				fWaitAll, dwMilliseconds);

		int intRtn = JInvoker.callIntProcAddress(lngFncAdrs[IDX_WaitForMultipleObjects], awfmo);
		
		System.out.println("WaitForMultipleObjects rtn: "  + Integer.toHexString(intRtn)
				+ " LastError: " + Integer.toString(awfmo.getLastError()));
		
		if (awfmo.getLastError() != 0) throw new Exception_Win64("WaitForMultipleObjects");
		if (intRtn == H_WinBase.WAIT_FAILED) throw new Exception_Win64("WaitForMultipleObjects");

		return intRtn;
	}
	
	/*
	BOOL CloseHandle(
	  HANDLE hObject
	);
	 */
	class Arg_CloseHandle extends Arg_property {
		private long hObject;
		
		Arg_CloseHandle(long hObject) {
			this.hObject = hObject;
		}
	}

	boolean CloseHandle(long hObject) throws Exception_Win64 {
		Arg_CloseHandle ach = new Arg_CloseHandle(hObject);

		boolean blRtn = JInvoker.callBooleanProcAddress(lngFncAdrs[IDX_CloseHandle], ach);
		
		System.out.println("CloseHandle ret: "  + Boolean.toString(blRtn)
				+ " LastError: " + Integer.toString(ach.getLastError()));
		
		if (ach.getLastError() != 0) throw new Exception_Win64("CloseHandle");
		if (blRtn == false) throw new Exception_Win64("CloseHandle");

		return blRtn;
	}
	
	/*
	HMODULE GetModuleHandle(
	  LPCTSTR lpModuleName   // W[
	);
	 */
	@_CharSet(CHARSET)
	class Arg_GetModuleHandle extends Arg_property {
		private @_Const String lpModuleName;

		Arg_GetModuleHandle(String lpModuleName) {
			this.lpModuleName = lpModuleName;
		}
	}
	
	/*
	HMODULE LoadLibrary(
	  LPCTSTR lpFileName   // W[̃t@C
	);
	 */
	@_CharSet(CHARSET)
	class Arg_LoadLibrary extends Arg_property {
		private @_Const String lpFileName;

		Arg_LoadLibrary(String lpFileName) {
			this.lpFileName = lpFileName;
		}
	}

	/*
	FARPROC GetProcAddress(
	  HMODULE hModule,    // DLL W[̃nh
	  LPCSTR lpProcName   // ֐
	);
	 */
	@_CharSet(Arg.ANSI)
	class Arg_GetProcAddress extends Arg_property {
		private long hModule;
		private @_Const @_Ansi String lpProcName;

		Arg_GetProcAddress(long hModule, String lpProcName) {
			this.hModule = hModule;
			this.lpProcName = lpProcName;
		}
	}
	
	/*
	BOOL FreeLibrary(
		HMODULE hModule   // DLL W[̃nh
	);
	 */
	class Arg_FreeLibrary extends Arg_property {
		private long hModule;

		Arg_FreeLibrary(long hModule) {
			this.hModule = hModule;
		}
	}

}
